// $Header: /home/owl/mackay/accel/split2/RCS/split2.cxx,v 1.2 1997/09/29 00:51:22 mackay Exp $
// Program: bacc.cxx -- simulates rf acceleration of a bunch of particles.
//  The program integrates difference equations based on the differential
//  equations 7.80 and 7.81 of  Conte and MacKay: "An Introduction to the
//  Physics of Particle Accelerators".


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <lagniappe.hxx>
#include <XWaldo.h>

#include "randists.h"

std::string cmdfile;
std::string initfile;

FILE *ave_fp = NULL;

const double EPSILON = 1.0e-29;	// tolerance for zero values.
const int NSTEP = 1;		// Number of integration steps per turn.

const int nlmax = 100;		// max number of input lines (This should
				//	 be replaced by a dynamic "new".
struct turndata
{
  int n;	// the number of turns for this input line.
  double ps0;	// the initial synch phase (degrees) for this input line.
  double psf;	// the final synch phase (degrees) for this input line.
  double rev0;	// the initial fraction of eV0max for this input line.
  double revf;	// the final fraction of eV0max for this input line.
  double rspin0;// the initial fraction of the spin kicks for this line.
  double rspinf;// the final fraction of the spin kicks for this line.
};
static struct turndata td[nlmax];

static int nl = 0;	// actual number of input lines read.

static int separatrix = 0;  // flag to draw separatix.
static int spin = 0;        // flag to indicate spin dependance.
static int autoflag = 0;    // =1 to auto-integrate.
static int ramp = 0;        // =1 to ramp ring with spin up energy.
static int two_d = 0;       // =1 for both (x,xp) and (phi,W) tracking
			    // =0 for just (phi,W) tracking
static int average_orbit=0; // =1 to calculate <x> and sigma_x, =0 to skip

// Some physics definitions (see function "init_rf")
const double pi = acos(-1.0);	        // Pi = 3.1415...
const double twopi = 2.0*pi;		// 2 Pi = 6.283...
const double c  = 2.99792458e+8;        // speed of light [m/s]
const double mp = 0.93827231e+9;        // proton mass [eV/c^2]
const double me = 0.51099906e+6;	// electron mass [eV/c^2]
const double mup = 1.4106e-26;		// magnetic moment of proton [J/T]
const double mue = 9.285e-24;		// magnetic moment of electron [J/T]
const double echarge = 1.60217733e-19;  // unit charge [C]
const double eps0 = 1.0e7/4.0/pi/c/c;   // [C/Vm] permitivity of free space
const double mAu = 197.0*0.93113e+9;	// gold mass [eV/c^2]
const double ZAu = 79;			// number of protons in Gold nucleus

static int iwave = 1;             // index for selected rf waveform
std::string wavetext[] =
  {
    "sin(phi)",
    "1st, 3rd harmonic approx to triangular wave",
    "1st, 3rd, 5th harmonic approx to triangular wave",
    "triangular wave",
    "sawtooth wave",
    "1st, 3rd harmonic with zero coeff for phi^3 term",
    "1st, 3rd, and 5th harms with no phi^3 or phi^5 terms"
  };

static double V;		  // rf voltage [V]
static double taus;		  // revolution period [s]
static double h;		  // rf harmonic number
static double omegarf;		  // rf frequency
static double gammat;		  // gamma (E/mc^2) at transition
static double gamma0;             // gamma (E/mc^2) of synchronous particle
static double beta;		  // v/c
static double Us;		  // energy of synchronous particle
static double eta;		  // phase slip factor
static double L;		  // circumference
static double eV0max;		  // max energy step per input line.
static double eV0;		  // actual energy step at beginning of line.
static double deV0;		  // energy step per turn.
static double ps;		  // synchronous phase.
static double sinps;              // sin(ps)
static double cosps;              // cos(ps)
static double dps;		  // synch phase step per turn.
static int il;			  // current line of data.
static int ilt;			  // current turn number for a given data line.
static double C1;                 // omegarf^2 * eta/(beta^3 * Us * c)
static double C2;                 // eV/Ctmp = eV/(L*omegarf) [eVs/m]
static double Ctmp;               // L*omegarf [m/s]
static double ds;		  // integration step [m]
static double Abk;		  // bucket area [eV s]
static double Wbk;                // height of stationary bucket [eV s]
static double Wspin;              // spin increment per turn
static double Gl = 1000.0;	  // G*l [T/m * m] for spin splitter amplitude
static double Wspin0;             // gamma * mu * G * l / omegarf [eVs]

static double dWspin;

static double Qs;                 // synchrotron tune
static double Wspinphase;         // phase of spin splitter modulation
static double dWspinphase;        // phase increment of splitter modulation
static double ratefactor = 1.0;   // fudge the modulation rate

static double bunchwidth;	  // halfwidth of bunch
static double Bf;                 // bunching factor
static double dpop0;              // initial delta p/p of bunch height
static double bunchheight;        // halfheight of bunch
static double Nb;                 // number of particles per bunch
static double g;                  // =1+2*ln(b/a); b/a is pipe radius/beam radius
static double Vb;                 // effective voltage of space charge.
static double momoffset = 0;	  // momentum offset for initial distribution
static double phioffset = 0;	  // phase offset of initial distribution

static double beta0;		  // initial beta0 of beam distribution  (x xp)
static double sqrtbeta;           // sqrt(beta0)
static double alpha0;             // initial alpha0 of beam distribution (x xp)
static double eta_x;		  // horiz dispersion
static double eta_xp;		  // derivative of horiz dispersion
static double epsilon0;		  // initial rms emittance/pi (x xp)
static double xoffset = 0;	  // initial horizontal offset of beam center
static double Q_x = 0.2;	  // horiz betatron tune (fractional or whole)
static double twopiQx = twopi*Q_x;
static double xi_x = -2.0;	  // horiz chromaticity
static double kappa = 0;	  // 2*Pi*alpha_j: horiz ampl tune-modulation 
static double C3;		  // -omegarf/(beta^2 * Us); dp/p = C3 * W
static double C3xi;		  // = C3*xi_x;

std::string comment;               // comment line from *.init file.

static double m, mu, buncharea0;

// Graphics information
int iframe = 0;
static double Xmin = -2.0*pi;  // graph limits
static double Xmax =  2.0*pi;
static double Ymin;
static double Ymax;
static int Ixmin, Ixmax;
static int Iymin, Iymax;
static int oleft, oright, otop, obottom;
double Wbk_sin;  // height of bucket with sin(phi) waveform
double xmin; // = -2.0*pi // limits currently being used
double xmax; // =  2.0*pi
double ymin; // = -Wbk_sin
double ymax; // =  Wbk_sin

static int xxp_plane = 0;    // set display: =0 for (phi,W); =1 for (x,xp)
static double x_max;         // (+/-)x limit for (x,xp) plots
static double xp_max;        // (+/-)xplimit for (x,xp) plots

// pen color numbers
//   black: 0, white: 1, red: 2, green: 3, yellow: 4,
//   cyan: 5, magenta: 6, orange3: 7
const int color_axes = 7;
const int color_labels = 1;
const int color_separatrix = 2;
const int color_spin_up = 5;
const int color_spin_down = 4;

const int LMARG = 10;
const int RMARG = 10;
const int TMARG = 10;
const int BMARG = 10;

// Arrays of coordinates of particles
struct particle
{
  double phi;
  double W;
  double x;
  double xp;
  double root_Ax;
  double psi_x;
};


static struct particle *bup;  // pointer to array of spin up particles.
static struct particle *bdn;  // pointer to array of spin down particles.

static int nbeam = 100;    // number of particles per spin.
//=================================
void helpit(const char* prognam)
{ 
  std::cerr << "Usage: "<<prognam <<" [-h] [-sep] [-f <rampfil>] ";
  std::cerr << "[-i <inifil>]\n";
  std::cerr << "  where -h           prints this message.\n";
  std::cerr << "        -sep         to draw separatrix.\n";
  // std::cerr << "        -spin        to add spin dependent calculation.\n";
  std::cerr << "        -f           sets the ramp input file to <rampfil>.\n";
  std::cerr << "                     The default name is 'ramp.cdr'.\n";
  std::cerr << "        -i           set the init file to <inifil>.\n";
  // std::cerr << "        -Gl          set a G*l for the peak spin cavity field.\n";
  std::cerr << "                     Defaults to 1000T\n";
  user_key(1, "h");

  std::cerr << "\nFormat of ramp file is 7 integers per line:\n";
  std::cerr << "\t\tn, ps0, psf, rev0, revf, rspin0, rspinf\n\n";
  std::cerr << "\tn\t number of turns to track (for this line)\n";
  std::cerr << "\tps0\t initial rf phase (degrees) for this line\n";
  std::cerr << "\tps1\t final rf phase for this line\n";
  std::cerr << "\trev0\t initial % of rf voltage for this line\n";
  std::cerr << "\trevf\t final % of rf voltage for this line\n";
  std::cerr << "\trspini\t initial % of strength of spin kicker for this line\n";
  std::cerr << "\trspinf\t final % of strength of spin kicker for this line\n";
  std::cerr << "\nThe f_rf, V_rf, and the spin kicker strength";
  std::cerr << " are ramped linearly for each line.\n";
}
/*===============================*/
int init_beam(void)
{ // initialize beam particles.
#ifdef REGULARBUNCHPATTERN
  int i, j;

  int circles = 4;
  int npercircle = 20;

//  int circles = 8;
//  int npercircle = 40;

  nbeam = circles*npercircle;
  iframe = 0;
  double dtheta = 2.0*pi/(double)npercircle;
  double theta;

  if(bup == NULL)bup = new struct particle[nbeam];

  double bw = 0;
  double bh = 0;
  double dbw = bunchwidth/(double)circles;
  double dbh = bunchheight/(double)circles;

  int k = 0;

  for(j=0;j<circles;j++)
    {
      theta = 0;
      bw += dbw;
      bh += dbh;
      for(i=0;i<npercircle;i++)
	{
	  bup[k].phi = bw*cos(theta);
	  bup[k].W   = bh*sin(theta);
	  theta += dtheta;
	  k++;
	}
    }

  if(spin)
    {
      if(bdn == NULL)bdn = new struct particle[nbeam];
      for(i=0;i<nbeam;i++)
	{
	  bdn[i].phi = bup[i].phi;
	  bdn[i].W = bup[i].W;
	}
    }
#else
// Use random distribution
  int i;
  double tmp;

  iframe = 0;
  srandom((unsigned int)1);

  if(bup == NULL)bup = new struct particle[nbeam];
  
  struct pt2d *xxp  = new struct pt2d[nbeam];
  struct pt2d *phiW = new struct pt2d[nbeam];

  if(arr_of_2dgauss(nbeam, bunchwidth/bunchheight,
			   0.0, bunchwidth*bunchheight/5.991, phiW))
    {
      std::cerr<<"arr_of_2dgauss(..., phiW) failed\n";
    }
  if(arr_of_2dgauss(nbeam, beta0, alpha0, epsilon0, xxp))
    {
      std::cerr<< "arr_of_2dgauss(..., xxp) failed\n";
    }
  double Woffset = momoffset/C3;
  for(i=0;i<nbeam;i++)
    {
      bup[i].phi = phiW[i].x + phioffset;
      bup[i].W   = phiW[i].y + Woffset;
      bup[i].x   = xxp[i].x + xoffset;
      bup[i].xp  = xxp[i].y;
      tmp = (alpha0*bup[i].x+beta0*bup[i].xp);

      bup[i].root_Ax = sqrt((bup[i].x*bup[i].x+tmp*tmp)/beta0);
      if(bup[i].root_Ax == 0)
	{
	  bup[i].psi_x = 0;
	}
      else
	{
	  bup[i].psi_x = acos(bup[i].x/bup[i].root_Ax/sqrtbeta);
	  if((beta0*bup[i].xp+alpha0*bup[i].x) < 0)
	    {
	      bup[i].psi_x = twopi - bup[i].psi_x;
	    }
	}
    }
  delete [] xxp;
  delete [] phiW;


  if(spin)
    {
      if(bdn == NULL)bdn = new struct particle[nbeam];
      for(i=0;i<nbeam;i++)
	{
	  bdn[i].phi = bup[i].phi;
	  bdn[i].W = bup[i].W;
	  bdn[i].x = bup[i].x;
	  bdn[i].xp = bup[i].xp;
	  bdn[i].root_Ax = bup[i].root_Ax;
	  bdn[i].psi_x = bup[i].psi_x;
	}
    }

#endif
}
/*===============================*/
int get_input_data(void)
{
  FILE *fh;
//  char line[100];
  
  fh = fopen((const char*)cmdfile.c_str(), "r");
  if(fh==NULL)
    {
      std::cerr <<"?Could not fopen("<<cmdfile<<")\n";
      return 1;
    }

  double ps0, psf;

  while(fscanf(fh, "%d %lf %lf %lf %lf %lf %lf",
	       &td[nl].n, &ps0, &psf,
	       &td[nl].rev0, &td[nl].revf,
	       &td[nl].rspin0, &td[nl].rspinf)!=EOF)
    {
      td[nl].ps0 = ps0*pi/180.;  // Convert from degrees to radians
      td[nl].psf = psf*pi/180.;
      if(nl++ >= nlmax)
	{
	  std::cerr <<"?Too many lines in file:"<<cmdfile<<"\n";
	  return -1;
	}
    }
  fclose(fh);
  fh = 0;
  if(nl <= 0)
    {
      std::cerr <<"Empty data file.\n";
      return -1;
    }
  return 0;
}
/*===============================*/
void dump_input_data(void)
{
  int i;
  for(i=0;i<nl;i++)
    {
      printf(" %5d %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
	     td[i].n, td[i].ps0, td[i].psf, td[i].rev0, td[i].revf,
	     td[i].rspin0, td[i].rspinf);
    }
}
/*===============================*/
int init_dataline(int linenum)
{
  if(linenum >= nl)
    {
      std::cerr<<"? init_dataline: linenum = "<<linenum<<" >= nl = "<<nl<<"\n";
      return 1;
    }

  struct turndata &tdp = td[linenum];
  eV0 = tdp.rev0*eV0max;
  Wspin0 = tdp.rspin0*Wspin;

  ps = tdp.ps0;
  sinps = sin(ps);
  cosps = cos(ps);

  if(tdp.n>1)
    {
      dps = (tdp.psf - tdp.ps0)/double(tdp.n);
      deV0 = eV0max*(tdp.revf - tdp.rev0)/double(tdp.n);
      dWspin = Wspin*(tdp.rspinf - tdp.rspin0)/double(tdp.n);
    } else {
      dps = 0;
      deV0 = 0;
      dWspin = 0;
    }
  ilt = 0; // reset turn counter to 0 for this line.
  return 0;
}
/*===============================*/
void dumpscaledvalue(FILE *fp, double value)
{
  double v = fabs(value);
  if(v<1000.)
    {
      fprintf(fp, "%f ", value);
    } else if(v<1.0e+6) {
      fprintf(fp, "%f k", value*0.001);
    } else if(v<1.0e+9) {
      fprintf(fp, "%f M", value*1.0e-6);
    } else if(v<1.0e+12) {
      fprintf(fp, "%f G", value*1.0e-9);
    } else if(v<1.0e+15) {
      fprintf(fp, "%f T", value*1.0e-12);
    } else {
      fprintf(fp, "%f P", value*1.0e-15);
    }
}
/*===============================*/
void dumpinit(FILE *fp, const char* prefix)
{
#define PFX fprintf(fp, "%s", prefix);
  PFX fprintf(fp, "%s\n", (const char*)comment.c_str());
  PFX fprintf(fp, "Initial parameter values for ");
  fprintf(fp, "Frame:%5d  l:%2d  t:%5d\n", iframe, il, ilt);
  PFX fprintf(fp, "bunching waveform: %s\n",
	      (const char*)wavetext[iwave-1].c_str());
  PFX fprintf(fp, "L=\t   %f m\n", L);
  PFX fprintf(fp, "m=\t   ");dumpscaledvalue(fp, m);fprintf(fp, "eV/c^2\n");
  PFX fprintf(fp, "mu=\t   %e J/T\n", mu);
  PFX fprintf(fp, "g=\t   %f\n", g);
  PFX fprintf(fp, "h=\t   %f\n", h);
  PFX fprintf(fp, "gammat=\t   %f\n", gammat);
  PFX fprintf(fp, "gamma0=\t   %f\n", gamma0);
  PFX fprintf(fp, "eta_tr=\t   %f\n", eta);
  PFX fprintf(fp, "beta=\t   %f\n", beta);
  PFX fprintf(fp, "Us=\t   ");dumpscaledvalue(fp, Us);fprintf(fp, "eV\n");
  PFX fprintf(fp, "beta_x=\t   %f m\n",beta0);
  PFX fprintf(fp, "alpha_x=   %f\n",alpha0);
  PFX fprintf(fp, "eta_x=\t   %f\n",eta_x);
  PFX fprintf(fp, "eta_xp=\t   %f\n",eta_xp);
  PFX fprintf(fp, "epsilon_x= %e m (rms)\n",epsilon0);
  PFX fprintf(fp, "Q_x=\t   %f\n",Q_x);
  PFX fprintf(fp, "xi_x=\t   %f   [=dQ/d(dp/p)]\n",xi_x);
  PFX fprintf(fp, "kappa=\t   %f\n",kappa);
  PFX fprintf(fp, "xoffset=   %f m\n",xoffset);
  PFX fprintf(fp, "momoffset= %f\n",momoffset);
  PFX fprintf(fp, "phioffset= %f\n",phioffset*180.0/pi);
  if(average_orbit)
    {
      PFX fprintf(fp, "ave_orb:   on\n");
    } else {
      PFX fprintf(fp, "ave_orb:   off\n");
    }
  
  if(spin)
    {
      PFX fprintf(fp,
		  "nbeam=\t   %d particles simulated (x2 for spin)\n", nbeam);
      PFX fprintf(fp, "Gl=\t   ");dumpscaledvalue(fp, Gl);fprintf(fp, "T\n");
      PFX fprintf(fp, "Wspin=\t   %e eV s\n", Wspin);
      if(ramp)
	{
	  PFX fprintf(fp, "ramp:\t   on\n");
	} else {
	  PFX fprintf(fp, "ramp:\t   off\n");
	}
    } else {
      PFX fprintf(fp, "nbeam=\t   %d particles simulated\n", nbeam);
    }
  PFX fprintf(fp, "Vrf=\t   ");dumpscaledvalue(fp, V);fprintf(fp, "V\n");
  PFX fprintf(fp, "Nb=\t   %e particles per bunch\n", Nb);
  PFX fprintf(fp, "h*Nb=\t   %e total particles\n", h*Nb);
  PFX fprintf(fp, "Vb=\t   %f V\n", Vb);
  PFX fprintf(fp, "Qs=\t   %f for a stationary bucket\n", Qs);
  PFX fprintf(fp, "1/Qs=\t   %f turns per synchrotron period\n", 1.0/Qs);
  PFX fprintf(fp, "f_sy=\t   %f Hz synchrotron frequency\n", Qs/taus);
  PFX fprintf(fp, "taus=\t   %e s\n", taus);
  PFX fprintf(fp, "frf=\t   ");
    dumpscaledvalue(fp, omegarf/2.0/pi);
    fprintf(fp, "Hz\n");

  PFX fprintf(fp, "dp/p=\t   %e (halfheight of initial bunch)\n", dpop0);
  PFX fprintf(fp,
	  "dphi=\t   %f deg (halfwidth of initial bunch)\n",
	  bunchwidth*180./pi);
  PFX fprintf(fp, "Bf=\t   %f (bunching factor)\n", Bf);
  PFX fprintf(fp, "Abk=\t   %e eV s  (area of stationary bucket)\n", Abk);
  PFX fprintf(fp, "Ab=\t   %e eV s  (initial bunch area)\n", buncharea0);
  PFX fprintf(fp, "Ab/Abk=\t   %f\n", buncharea0/Abk);
}
/*===============================*/
int init_rf(void)
{
  std::string Tline;
  const int LENLINE = 100;
  char line[LENLINE];
  char* what;

  std::ifstream ifl(initfile.c_str());
  if(! ifl)
    {
      std::cerr << "Cannot open \""<<initfile<<"\" for input.\n";
      helpit("<program>");
      exit(1);
    }

// Put in some default values in case something is missing.
  L = 3833.845;			// m
  V = 300000.0;			// V
  h = 342;			// harmonic number for for acceleration
  gammat = 22.8;
  gamma0 = 31.2;
  mu = mup;
  m = mp;

  while(! ifl.eof())
    {
      //      ifl >> Tline;
      std::getline(ifl, Tline);
      if(Tline.size()>0)
	{
	  strncpy(line, (const char*)Tline.c_str(), LENLINE);
	  what = strtok(line, ": \t");
	  if(strcmp(what, "particle")==0)
	    {
	      what = strtok(NULL, ": \t");
	      if(strcmp(what, "electron")==0)
		{
		  m = me;
		  mu = mue;
		}
	      else if(strcmp(what, "proton")==0)
		{
		  m = mp;
		  mu = mup;
		}
	      else if(strcmp(what, "Au")==0)
		{
		  m = mAu/double(ZAu);	// effective mass/charge
		  mu = 0.;		//Ignore polarization for gold.
		}
	    }
	  else if(strcmp(what, "L")==0)
	    {
	      what = strtok(NULL, ": \t");
	      L = atof(what);
	    }
	  else if(strcmp(what, "Vrf")==0)
	    {
	      what = strtok(NULL, ": \t");
	      V = atof(what);
	    }
	  else if(strcmp(what, "h")==0)
	    {
	      what = strtok(NULL, ": \t");
	      h = atof(what);
	    }
	  else if(strcmp(what, "gammat")==0)
	    {
	      what = strtok(NULL, ": \t");
	      gammat = atof(what);
	    }
	  else if(strcmp(what, "gamma0")==0)
	    {
	      what = strtok(NULL, ": \t");
	      gamma0 = atof(what);
	    }
	  else if(strcmp(what, "bunchwidth")==0)
	    {
	      what = strtok(NULL, ": \t");
	      bunchwidth = atof(what)*pi/180.;
	    }
	  else if(strcmp(what, "epsilon")==0)
	    {
	      what = strtok(NULL, ": \t");
	      epsilon0 = atof(what)/5.991;
	    }
	  else if(strcmp(what, "beta")==0)
	    {
	      what = strtok(NULL, ": \t");
	      beta0 = atof(what);
	      sqrtbeta = sqrt(beta0);
	    }
	  else if(strcmp(what, "alpha")==0)
	    {
	      what = strtok(NULL, ": \t");
	      alpha0 = atof(what);
	    }
	  else if(strcmp(what, "eta_x")==0)
	    {
	      what = strtok(NULL, ": \t");
	      eta_x = atof(what);
	      if(eta_x != 0) std::cerr<< "Ignoring non-zero dispersion (eta_x)\n";
	    }
	  else if(strcmp(what, "eta_xp")==0)
	    {
	      what = strtok(NULL, ": \t");
	      eta_xp = atof(what);
	      if(eta_xp != 0) std::cerr<< "Ignoring non-zero dispersion (eta_xp)\n";
	    }
	  else if(strcmp(what, "Q_x")==0)
	    {
	      what = strtok(NULL, ": \t");
	      Q_x = atof(what);
	      twopiQx = twopi*Q_x;
	    }
	  else if(strcmp(what, "xi_x")==0)
	    {
	      what = strtok(NULL, ": \t");
	      xi_x = atof(what);
	    }
	  else if(strcmp(what, "kappa")==0)
	    {
	      what = strtok(NULL, ": \t");
	      kappa = atof(what);
	    }
	  else if(strcmp(what, "xoffset")==0)
	    {
	      what = strtok(NULL, ": \t");
	      xoffset = atof(what);
	    }
	  else if(strcmp(what, "momoffset")==0)
	    {
	      what = strtok(NULL, ": \t");
	      momoffset = atof(what);
	    }
	  else if(strcmp(what, "phioffset")==0)
	    {
	      what = strtok(NULL, ": \t");
	      phioffset = atof(what)*pi/180.0;
	    }
	  else if(strcmp(what, "dp/p")==0)
	    {
	      what = strtok(NULL, ": \t");
	      dpop0 = atof(what);
	    }
	  else if(strcmp(what, "Nb")==0)
	    {
	      what = strtok(NULL, ": \t");
	      Nb = atof(what);
	    }
	  else if(strcmp(what, "g")==0)
	    {
	      what = strtok(NULL, ": \t");
	      g = atof(what);
	    }
	  else if(strcmp(what, "ratefactor")==0)
	    {
	      what = strtok(NULL, ": \t");
	      ratefactor = atof(what);
	    }
	  else if(strcmp(what, "ramp")==0)
	    {
	      what = strtok(NULL, ": \t");
	      if(strcmp(what, "on")==0)ramp = 1;
	    }
	  else if(strcmp(what, "two_d")==0)
	    {
	      what = strtok(NULL, ": \t");
	      if(strcmp(what, "yes")==0)two_d = 1;
	    }
	  else if(strcmp(what, "x_limit")==0)
	    {
	      what = strtok(NULL, ": \t");
	      x_max = atof(what);
	    }
	  else if(strcmp(what, "xp_limit")==0)
	    {
	      what = strtok(NULL, ": \t");
	      xp_max = atof(what);
	    }
	  else if(strcmp(what, "ave_orb")==0)
	    {
	      what = strtok(NULL, ": \t");
	      if(strcmp(what, "yes")==0)average_orbit = 1;
	    }
	  else if(strcmp(what, "nbeam")==0)
	    {
	      what = strtok(NULL, ": \t");
	      nbeam = atoi(what);
	    }
	  else if(strcmp(what, "comment")==0)
	    {
	      comment = Tline;
	    }
	  else
	    {
	      if(Tline[0]!='#')	std::cout << "Weird line in \"" << initfile
		                     << "\": " << Tline << "\n";
	    }
	}
      else
	{
//	  std::cout << "Tline.size()<=0\n";
	}
    }

  Us = m*gamma0;		// eV
  beta = sqrt(1.0 - 1.0/gamma0/gamma0);

  taus = L/beta/c;		// s
  omegarf = h*2*pi/taus;	// radian/s
  eta = 1.0/gamma0/gamma0 - 1.0/gammat/gammat;
  Bf = bunchwidth/pi;
  Vb = 6*g*Nb*h*h*echarge/(pi*gamma0*gamma0*Bf*Bf*Bf*4.0*pi*eps0*L);

  eV0max = V - Vb;  // V

  Wspin = mu/echarge/omegarf;   // mu /omegarf  [e m^2]
  Wspin = gamma0*Wspin*Gl/L;    // [eV s/m]

  C1 = omegarf*omegarf*eta/(beta*beta*beta*Us*c);
  Ctmp = L*omegarf;
  C2 = eV0max/2.0/pi/h/beta/c;

  C3 = -omegarf/(beta*beta*Us);  // conversion: delta = dp/p = C3 * W
  C3xi = C3*xi_x/Q_x;

  ds = L/double(NSTEP);
// from Eq. 7.103 of Conte & MacKay:
//      Abk = (8.0*L/pi/c)*sqrt(eV0max*Us/2.0/pi/h/h/h/eta); // eV s
  Abk = 16.0*sqrt(C2/C1);  // ev s
  Wbk_sin = 2.*sqrt(C2/C1);
  switch (iwave)
    {
    case 1:  // sin(phi) waveform
      Wbk = Wbk_sin;
      break;
    case 2:  // 1st, 3rd harmonic approx to triangular wave
      Abk *= 1.095;
      Wbk = sqrt((16.0*C2/C1)*(1.0-1.0/27.0));
      break;
    case 3:  // 1st, 3rd, 5th harmonic approx to triangular wave
      Abk *= 1.104;
      Wbk = sqrt((16.0*C2/C1)*(1.0-1.0/27.0+1.0/125.0));
      break;
    case 4:  // triangular wave
      Abk *= 1.101;
      Wbk = sqrt((16.0*C2/C1)*0.968946);
      break;
    case 5:  // sawtooth wave
      Abk = pi*pi*pi*sqrt(C2/C1);
      Wbk = pi*sqrt(C2/C1);
      break;
    case 6:  // (9/8)sin(x) - sin(3x)/24
      Wbk = Wbk_sin;
      break;
    case 7:  // harmonics 1, 3, 5 with no phi^3 or phi^5 term
      Wbk = Wbk_sin;
      break;
    }      
  
  if(xxp_plane)
    {
      xmax = x_max;
      xmin = -xmax;
      ymax = xp_max;
      ymin = -ymin;
    } else {
      xmin = -2.0*pi;
      xmax = 2.0*pi;
      ymin = -Wbk_sin;
      ymax = Wbk_sin;
    }

  bunchheight = dpop0*Us*beta*beta/omegarf;
  buncharea0 = pi*bunchheight*bunchwidth;

  Qs = sqrt(h*eta*eV0max/2.0/pi/beta/beta/gamma0/m);
  Wspinphase = 0;
  dWspinphase = 2.0*pi*Qs*ratefactor;

  dumpinit(stdout, "");

  il = 0;
  return init_dataline(il);
}
/*===============================*/
double sepatrix(double phi, double ps)
{
  double tmp;
  double a = 2.*C2/C1;
  if(a<=0.0)return 0.0; // give up for now.

  switch (iwave)
    {
     // these look the same, so I will cheat by using the 1st and 3rd harmonics
    case 2:  // 1st and 3rd harmonic approx to triangular waveform
      return sqrt((a*4.0/pi)*(1.0+cos(phi) - (1.0+cos(3*phi))/27.0));
      break;
    case 3:  // 1st, 3rd, 5th harmonic approx to triangular waveform
    case 4:  // triangular waveform
      return sqrt((a*4.0/pi)*(1.0+cos(phi)
			      - (1.0+cos(3*phi))/27.0
			      + (1.0+cos(5*phi))/125.0));
      break;
    case 5:  // sawtooth waveform
      tmp = a*(pi*pi-phi*phi)/2.0;
      if(tmp<0)return 0;
      return sqrt(tmp);
      break;

    case 1:  // sin(phi)
    default:
      if(ps==0)
	{
	  return sqrt(a*(1.0+cos(phi)));
	}else{
	  if(eta>0)
	    {
	      double c = cos(ps);
	      double t = tan(ps);
	      tmp = (cos(phi)-cos(pi-ps))/c + (phi-pi+ps)*t;
	      if(tmp<0)tmp = 0;
	      return sqrt(a*tmp);
	    }else{
	      double c = cos(ps);
	      double t = tan(ps);
	      tmp = (cos(phi)-cos(-pi-ps))/c + (phi+pi+ps)*t;
	      if(tmp<0)tmp = 0;
	      return sqrt(a*tmp);
	    }
	}
    }
  
}
/*===============================*/
int draw_separatrix(double ps)
{   // just try to draw separatrices for stationary buckets
  double phi, w;
  int nsteps = 50;
  double dphi = 2.0*pi/double(nsteps);
  double i, j;

  setpen(color_separatrix);
  for(j=0;j<2;j++)
    {
      phi = -pi;
      for(i=0;i<nsteps;i++)
	{
	  if(i==0)movea(phi, 0.);
	  phi += dphi;
	  w = sepatrix(phi,ps);
	  if(w<EPSILON)w = 0;
	  if(j==0)
	    {
	      drawa(phi, w); // top half
	    } else {
	      drawa(phi, -w); // bottom half
	    }
	}
    }
}
/*===============================*/
int dump_separatrix(double ps)
{  // write out a file with the current separatrix for gnuplot
  FILE *sfp = fopen("separatrix.out", "w");
  const double nsteps = 50.0;
  int i, j;
  double theta;
  double sign;

  fprintf(sfp, "# separatrix for waveform %d\n", iwave);
  fprintf(sfp, "%f\t%e\n", -pi, 0.0);
  double dtheta = 2.0*pi/nsteps;
  theta = -pi;

  for(j=0;j<2;j++)
    {
      if(j==1)
	{
	  dtheta = -dtheta;
	  sign = -1.0;
	} else {
	  sign = 1.0;
	}
      for(i=0;i<nsteps;i++)
	{
	  theta +=dtheta;
	  fprintf(sfp, "%f\t%e\n", theta, sign*sepatrix(theta,ps));
	}
    }
  fclose(sfp);
  sfp = NULL;
}
/*===============================*/
/*===============================*/
/*===============================*/
void user_start(int argc, char** argv)
{
//  init_rf();
  init_beam();
}
/*===============================*/
void user_mouse(int mbutton, int xmdown, int ymdown, int xmup, int ymup)
{
  if(mbutton==1)
    { // zoom in by a factor of two
      xmin /= 2.0;
      xmax /= 2.0;
      ymin /= 2.0;
      ymax /= 2.0;
    } else if(mbutton==2) {
      xmin *= 2.0;
      xmax *= 2.0;
      ymin *= 2.0;
      ymax *= 2.0;
    } else {
      if(xxp_plane)
	{
	  xmax = x_max;
	  xmin = -xmax;
	  ymax = xp_max;
	  ymin = -ymax;
	}
      else
	{
	  xmin = -2.0*pi;
	  xmax = 2.0*pi;
	  ymin = -Wbk_sin;
	  ymax = Wbk_sin;
	}
    }
  erase();
  user_refresh(oleft, oright, otop, obottom);
}
/*===============================*/
void user_refresh(int left, int right, int top, int bottom)
{
  int i;
  char lbl[80];

  oleft = left;
  oright = right;
  otop = top;
  obottom = bottom;

  setregion(left+LMARG, right-RMARG, top+TMARG, bottom-BMARG);
  double xfudge = 1.0;
  double yfudge = 2.0;
  setvirtual(xmin*xfudge, xmax*xfudge, ymin*yfudge, ymax*yfudge);
//  setvirtual(2*xmin, 2*xmax, 2*ymin, 2*ymax);
  clip(1);
  setpen(color_axes);
  movea(xmin, 0.);
  drawa(xmax, 0.);
  movea(0., ymin);
  drawa(0., ymax);
  
  if(!xxp_plane)
    {
      C2 = eV0/Ctmp;
      // Plot separatrix
      if(separatrix)
	{
	  draw_separatrix(ps);
	}

      // Plot particles
      setpen(color_spin_up);
      for(i=0;i<nbeam;i++)
	{
	  movea(bup[i].phi, bup[i].W);
	  drawa(bup[i].phi, bup[i].W);
	}
      if(spin)
	{
	  setpen(color_spin_down);
	  for(i=0;i<nbeam;i++)
	    {
	      movea(bdn[i].phi, bdn[i].W);
	      drawa(bdn[i].phi, bdn[i].W);
	    }
	}
    }
  else
    {
      // Plot particles
      setpen(color_spin_up);
      for(i=0;i<nbeam;i++)
	{
	  movea(bup[i].x, bup[i].xp);
	  drawa(bup[i].x, bup[i].xp);
	}
      if(spin)
	{
	  setpen(color_spin_down);
	  for(i=0;i<nbeam;i++)
	    {
	      movea(bdn[i].x, bdn[i].xp);
	      drawa(bdn[i].x, bdn[i].xp);
	    }
	}
    }
  
  clip(0);
  sprintf(lbl, "Frame:%5d  l:%2d  t:%5d", iframe, il, ilt);
  setpen(color_labels);
  lblabs(1, 1, lbl, 3);
  clip(1);
}
/*===============================*/
double waveform(double phi)
{

  switch (iwave)
    {
    case 1:   // first harmonic sine wave cavity
      return sin(phi);
    case 2:   // first and third harmonic cavities
      return (4.0/pi)*(sin(phi)-sin(3*phi)/9.0); // approximates a triangle
    case 3: // approximates a triangle with 1st, 3rd, and 5th harmonics
      return (4.0/pi)*(sin(phi)-sin(3*phi)/9.0+sin(5*phi)/25.0);
    case 4:   // triangular wave
      if(phi<-1.5*pi)
	{
	  return phi+2.0*pi;
	}
      else if(phi<-0.5*pi)
	{
	  return -pi-phi;
	}
      else if(phi<0.5*pi)
	{
	  return phi;
	}
      else if(phi<1.5*pi)
	{
	  return pi-phi;
	}
      else
	{
	  return phi-2.0*pi;
	}
    case 5:   // sawtooth (good from -3pi --> +3pi)
      if(phi<-pi)
	{
	  return phi+2.0*pi;
	}
      else if (phi<pi)
	{
	  return phi;
	}
      else
	{
	  return phi-2.0*pi;
	}
    case 6:   // 1st and 3rd harmonics with phi^3 coeff = 0
      {
	return (9.0/8.0)*sin(phi) - sin(3.0*phi)/24.0;
      }
    case 7:   // 1st, 3rd and 5th harmonics with phi^3 and phi^5 coeffs = 0
      {
	return (75.0*sin(phi)-25.0*sin(3.0*phi)/6.0+0.3*sin(5.0*phi))/64.0;
      }
    }
}
/*===============================*/
void do_average_orbit(void)
{
  double sxup, sxup2, aveup, sigmaup;
  double sxdn, sxdn2, avedn, sigmadn;
  int i;

  if(ave_fp == NULL)
    {
      ave_fp = fopen("ave_orbit.out", "w");
      if(ave_fp == NULL)
	{
	  std::cerr<<"?Could not fopen(ave_orbit.out)\n";
	  exit(1);
	}
    }

  sxup = sxup2 = sxdn = sxdn2 = 0;
  for(i=0;i<nbeam;i++)
    {
      sxup += bup[i].x;
      sxup2 += bup[i].x*bup[i].x;
      if(spin)
	{
	  sxdn += bdn[i].x;
	  sxdn2 += bdn[i].x*bdn[i].x;
	}
    }
  double n = nbeam;
  aveup = sxup/n;
  sigmaup = sqrt(sxup2/n - aveup*aveup);
  if(spin)
    {
      avedn = sxdn/n;
      sigmadn = sqrt(sxdn2/n - avedn*avedn);
      fprintf(ave_fp, "%e\t%e\t%e\t%e\n", aveup, sigmaup, avedn, sigmadn);
    } else {
      fprintf(ave_fp, "%e\t%e\n", aveup, sigmaup);
    }
}
/*===============================*/
int integrate(void)
{
  int i, j;
  double phin;
  double tmpWspin;
  double cospsi;
  double wfps;

  ilt++;
  if(ilt >= td[il].n)
    {
      ilt--;
      il++;
      if(il >= nl)  // Have we reached the last frame?
	{
	  il = nl-1;
	  autoflag = 0;
	  return 1;
	}
      init_dataline(il);
    } else {
      ps += dps;
      eV0 += deV0;
      Wspin0 += dWspin;
    }

  iframe++;

//  sinps = sin(ps);
//  cosps = cos(ps);

  C2 = eV0/Ctmp;

  if(spin)tmpWspin = sin(Wspinphase)*Wspin0*ds;
  wfps = waveform(ps);
  for(i=0;i<nbeam;i++)
    {
      for(j=0;j<NSTEP;j++)
	{
	  bup[i].W -= C2*(waveform(bup[i].phi)-wfps)*ds;
	  bup[i].phi = bup[i].phi + C1*bup[i].W*ds;
	  if(two_d)
	    {
	      bup[i].psi_x += twopiQx*(1.0+C3xi*bup[i].W)
		+ kappa*bup[i].root_Ax*bup[i].root_Ax;
	      cospsi = cos(bup[i].psi_x);
	      bup[i].x = bup[i].root_Ax*sqrtbeta*cospsi;
	      bup[i].xp = (bup[i].root_Ax/sqrtbeta)
		*(alpha0*cospsi+sin(bup[i].psi_x));
	    }
	  if(spin)
	    {
	      bdn[i].W -= C2*waveform(bdn[i].phi)*ds;
	      bdn[i].phi = bdn[i].phi + C1*bdn[i].W*ds;
	      if(two_d)
		{
		  bdn[i].psi_x += twopiQx*(1.0+C3xi*bdn[i].W)
		    + kappa*bdn[i].root_Ax*bdn[i].root_Ax;
		  cospsi = cos(bdn[i].psi_x);
		  bdn[i].x = bdn[i].root_Ax*sqrtbeta*cospsi;
		  bdn[i].xp = (bdn[i].root_Ax/sqrtbeta)
	    			    *(alpha0*cospsi+sin(bdn[i].psi_x));
		}
	    }
	}
      if(spin)
	{
	  if(ramp)
	    {
	      bdn[i].W -= 2*tmpWspin;
	    } else {
	      bup[i].W += tmpWspin;
	      bdn[i].W -= tmpWspin;
	    }
	}
    }
  if(spin)Wspinphase += dWspinphase;

  if(average_orbit)do_average_orbit();

  return 0;
}
/*===============================*/
void dumpvalues(void)
{
  FILE *fp = fopen("baccsr.out", "w");
  int i;

  dumpinit(fp, "#  ");

  if(spin)
    {
      for(i=0;i<nbeam;i++)
	{
	  fprintf(fp, "%e %e %e %e %e %e %e %e\n",
		  bup[i].phi, bup[i].W, bup[i].x, bup[i].xp,
		  bdn[i].phi, bdn[i].W, bdn[i].x, bdn[i].xp);
	}
    }
  else
    {
      for(i=0;i<nbeam;i++)
	{
	  fprintf(fp, "%e %e %e %e\n",
		  bup[i].phi, bup[i].W,
		  bup[i].x, bup[i].xp);
	}
    }

  fclose(fp);
}
/*===============================*/
void flush_aveorb(void)
{
  if(ave_fp)fflush(ave_fp);
}
/*===============================*/
void user_key(int i, char *code)
{
  double refresh = 0;
  int j;

  switch (code[0])
    {
    case 'h':
      std::cerr<<"Keyboard input:\n";
      std::cerr<<"\th: type this message.\n";
      std::cerr<<"\ti: reinitialize the beam particles.\n";
      std::cerr<<"\tn: move particles to next frame.\n";
      std::cerr<<"\ta: automatically propagate.\n";
      std::cerr<<"\tA: automatically propagate without diplaying until done.\n";
      std::cerr<<"\tp: propagate through 1 synch period\n";
      std::cerr<<"\tP: propagate through 1 synch period without";
         std::cerr<<" displaying until done\n";
      std::cerr<<"\tr: refresh the plot.\n";
      std::cerr<<"\ts: toggle plot of separatrix.\n";
      std::cerr<<"\tw: write out current turn's data to file.\n";
      std::cerr<<"\tx: change display to (x,xp) plane if 'two_d' is true\n";
      std::cerr<<"\tz: change display to (phi,W) plane if 'two_d' is true\n";
      std::cerr<<"\tq: quit\n";
      break;

    case 's':
      if(separatrix == 0)
	{
	  separatrix = 1;
	} else {
	  separatrix = 0;
	}
      refresh = 1;
      break;

    case 'r':
      refresh = 1;
      break;

    case 'i':
      nl = 0;
      get_input_data();
      init_rf();
      init_beam();
      refresh = 1;
      if(ave_fp)  // flush the old file and restart
	{
	  fclose(ave_fp);
	  ave_fp = NULL;
	}
      if(xxp_plane)
	{
	  xmax = x_max;
	  xmin = -xmax;
	  ymax = xp_max;
	  ymin = -ymax;
	}
      break;

    case 'n':
      integrate();
      flush_aveorb();
      refresh = 1;
      break;

    case 'w':
      dumpvalues();
      dump_separatrix(0.0);
      break;

    case 'A':
      autoflag = 1;
      while(autoflag)
	{
	  integrate();
	}
      flush_aveorb();
      refresh = 1;
      break;

    case 'a':
      autoflag = 1;
      while(autoflag)
	{
	  integrate();
	  erase();
	  user_refresh(oleft, oright, otop, obottom);
	}
      flush_aveorb();
      break;

    case 'p':
      j =1;
      while(j<1.0/Qs)
	{
	  integrate();
	  erase();
	  user_refresh(oleft, oright, otop, obottom);
	  j++;
	}
      flush_aveorb();
      break;

    case 'P':
      j =1;
      while(j<1.0/Qs)
	{
	  integrate();
	  j++;
	}
      flush_aveorb();
      refresh = 1;
      break;

    case 'x':
      if(two_d == 0)
	{
	  std::cerr<<"Not running with 2 degrees of freedom: update 'bacc.init'\n";
	  xxp_plane = 0;
	} else {
	  xxp_plane = 1;
	  xmax = x_max;
	  xmin = -xmax;
	  ymax = xp_max;
	  ymin = -ymax;
	}
      refresh = 1;
      break;
    case 'z':
      xxp_plane = 0;
      refresh = 1;
      xmin = -2.0*pi;
      xmax = 2.0*pi;
      ymin = -Wbk_sin;
      ymax = Wbk_sin;
      break;

    case '1':
      iwave = 1;
      std::cout<<"Set wavform using 1st harmonic cavity\n";
      if(separatrix) refresh = 1;
      break;

    case '2':
      iwave = 2;
      std::cout<<"Set wavform using 1st and 3rd harmonic cavities\n";
      if(separatrix) refresh = 1;
      break;

    case '3':
      iwave = 3;
      std::cout<<"Set waveform using 1st, 3rd, and 5th harmonic cavities\n";
      if(separatrix) refresh = 1;
      break;

    case '4':
      iwave = 4;
      std::cout<<"Set waveform using triangular wave\n";
      if(separatrix) refresh = 1;
      break;

    case '5':
      iwave = 5;
      std::cout<<"Set waveform using sawtooth wave\n";
      if(separatrix) refresh = 1;
      break;

    case '6':
      iwave = 6;
      std::cout<<"Set waveform using 1st and 3rd harmonic with no phi^3 term\n";
      if(separatrix) refresh = 1;
      break;

    case '7':
      iwave = 7;
      std::cout<<"Set waveform using 1st, 3rd, and 5th harms with no phi^3 or phi^5 terms\n";
      if(separatrix) refresh = 1;
      break;
    }
  if(refresh)
    {
      erase();
      user_refresh(oleft, oright, otop, obottom);
    }
}
/*===============================*/
int process_cmd(int *argc, char** argv)
{
  if(eat_switch(argc, argv, "-h"))
    {
    //  std::cerr << "Usage: "<<argv[0]<<" [-h] [-spin] [-sep] [-f <rampfil>] ";
    //  std::cerr << "[-i <inifil>] [-Gl <fudge>]\n";

      helpit(argv[0]);
      return 1;
    }

  separatrix = eat_switch(argc, argv, "-sep");
  spin = eat_switch(argc, argv, "-spin");

  cmdfile = eat_switch_with_arg(argc, argv, "-f");
  if(cmdfile.size() == 0)
    {
      cmdfile = "ramp.cdr";
    }

  initfile = eat_switch_with_arg(argc, argv, "-i");
  if(initfile.size() == 0)
    {
      initfile = "bacc.init";
    }

  std::string fudge = eat_switch_with_arg(argc, argv, "-Gl");
  if(fudge.size() != 0)
    {
      Gl = atof((const char*)fudge.c_str());
    }

  return 0;
}
/*===============================*/
int main(int argc, char** argv)
{
  if(process_cmd(&argc, argv))return 1;
  if(get_input_data())return 1;
  if(init_rf())return 1;
  dump_input_data();

  XWaldoEventLoop(argc, argv);
  return 0;
}
