// su2mat.hxx
#ifndef SU2MAT_HXX
#define SU2MAT_HXX 1

#include <complex>
#include <string>
#include <stdio.h>
#include <string.h>

#include <fstream>
#ifndef SU2MAT_FORMAT
#define SU2MAT_FORMAT " %11.8lf %+11.8lf i "
#endif

std::complex<double> I(0.0,1.0);

//=========================================
//external char* su2mat_format;
//char su2mat_format[]=" %11.8lf %+11.8lf i";

class su2mat
{
public:
  std::complex<double> m[2][2];
  //-----------
  su2mat()
  {
    m[0][0]=m[1][1] = 1.0;
    m[1][0]=m[0][1] = 0.0;
  }

  su2mat operator+(const su2mat &a) const
  {// This just adds two matrices and the result in not in SU2
    int i, j;
    su2mat b;

    for(i=0;i<2;i++)
      {
	for(j=0;j<2;j++)
	  {
	    b.m[i][j] = m[i][j]+a.m[i][j];
	  }
      }
    return b;
  }
  // -  -  -  -
  su2mat operator-(const su2mat &a) const
  {// This just subtracts two matrices and the result in not in SU2
    int i, j;
    su2mat b;

    for(i=0;i<2;i++)
      {
	for(j=0;j<2;j++)
	  {
	    b.m[i][j] = m[i][j]-a.m[i][j];
	  }
      }
    return b;
  }
  // -  -  -  -
  su2mat operator*(const su2mat &a)const
  {
    int i, j;
    su2mat b;
    for(i=0;i<2;i++)
      {
	for(j=0;j<2;j++)
	  {
	    b.m[i][j]=m[i][0]*a.m[0][j]+m[i][1]*a.m[1][j];
	  }
      }
    return b;
  }
  // -  -  -  -
  su2mat unit()
  {
    int i, j;
    for(i=0;i<2;i++)
      {
	for(j=0;j<2;j++)
	  {
	    if(i==j)m[i][j] = 1.0;
	    else m[i][j] = 0.0;
	  }
      }
    return *this;
  }
  // -  -  -  -
  friend su2mat unit(void)
  {
    su2mat M;
    int i, j;
    M.m[0][0] = 1;
    M.m[1][1] = 1;
    return M;
  }
  // -  -  -  -
  friend su2mat transpose(const su2mat N)
  {
    su2mat M;
    int i, j;
    std::complex<double> x;

    for(i=0;i<2;i++)
      {
	for(j=i+1;j<2;j++)
	  {
	    x = N.m[i][j];
	    M.m[i][j] = N.m[j][i];
	    M.m[j][i] = x;
	  }
      }
    return M;
  }
  // -  -  -  -
  friend su2mat inv(const su2mat N)
  {
    su2mat M;
    int i, j;
    std::complex<double> x;

    for(i=0;i<2;i++)
      {
	M.m[i][i] = std::conj(N.m[i][i]);
	for(j=i+1;j<2;j++)
	  {
	    x = N.m[i][j];
	    M.m[i][j] = std::conj(N.m[j][i]);
	    M.m[j][i] = std::conj(x);
	  }
      }
    return M;
  }
  // -  -  -  -
  su2mat &operator=(const su2mat &a);
  // -  -  -  -
  friend std::ostream& operator<<(std::ostream &os, const su2mat &v)
  {  // This does gives a segmentation fault
    int i, j;
    char numstr[50];
    for(i=0;i<2;i++)
      {
	for(j=0;j<2;j++)
	  {
	    sprintf(numstr, SU2MAT_FORMAT,
		    real(v.m[i][j]), imag(v.m[i][j]));
	    os << "\t " << numstr;
	  }
	os << "\n";
      }
    return os;
  }
};
//=================
su2mat &su2mat::operator=(const su2mat &a)
{
  int i, j;
  for(i=0;i<2;i++)
    {
      for(j=0;j<2;j++)
	{
	  m[i][j]=a.m[i][j];
	}
    }
  return *this;
}
// -  -  -  -
su2mat Rx(const double mu)
{
  su2mat M;
  M.m[0][0] = cos(mu/2);
  M.m[1][1] = M.m[0][0];
  M.m[0][1] = I*sin(mu/2);
  M.m[1][0] = M.m[0][1];
  return M;
}
// -  -  -  -
su2mat Ry(const double mu)
{
  su2mat M;
  M.m[0][0] = cos(mu/2);
  M.m[1][1] = M.m[0][0];
  M.m[0][1] = sin(mu/2);
  M.m[1][0] = -M.m[0][1];
  return M;
}
// -  -  -  -
su2mat Rz(const double mu)
{
  su2mat M;
  M.m[0][0] = exp(mu*I/2.0);
  M.m[1][1] = conj(M.m[0][0]);
  return M;
}
// -  -  -  -
su2mat Rn(const double mu, Point n)
{ // note: "n" must be normalized.
  su2mat R;
  double cs, sn;
  std::complex<double> a, b;
  cs = cos(mu/2.);
  sn = sin(mu/2.);
  a = cs+I*n.z*sn;
  b = (n.y+I*n.x)*sn;
  R.m[0][0] = a;
  R.m[0][1] = b;
  R.m[1][0] = -std::conj(b);
  R.m[1][1] = std::conj(a);
  return R;
}

 
#endif // SU2MAT_HXX
