// $Id: $
// Program: blspin.cxx  (nonlinear version under development).
// Particle tracking program based on the Tfunc class from the old bl program.
// The idea is to get rid of the old database driven stuff and just input
// the lattice from a simple flat file.  I plan to add spin tracking as well,
// although it should be without acceleration, at least at first.
// This version of blspin hass been modified to calculate stroboscopic averages
// of the stable spin field.

// 8 September, 2010:  hopefully fixed element offsets for linear elements.
//   Notes: Do not apply offsets to the steering correctors, otherwise
//     any bumps will be tromped on.
//   At present nonlinear elements are initialized to unit matrices with
//     with any offsets being lost since v=(I-M)*voff.

// Author: W. W. MacKay
// Bugs: of course.
//---------------------

#include <iostream>
#include <string>
#include <sstream>

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Tfunc.hxx"
#include "nl_blspin_gdef.inc"
#include "su2/su2spinor.hxx"
#include "mat3/mat3.hxx"
// -- --
const int MAXLIN = 400;

const double pi = acos(-1.0);
const double twopi = 2.0*pi;

// Define a few parameters with some defaults to be changed later.
double circumference;
double alpha_p = 0;     // initialize to zero so that things don't break.
double gam = 100.0;     // Lorentz factor.
double beta;            // v/c for design particle.
double Gg;              // G*gam for spin precession
double omegarf;         // 2pi times rf frequency
double Ggp1;		// G*gamma+1
double Ggm1;		// G*(gamma-1)
double Gp1;		// G+1
double rigidity;	// p_0/q

int nlines = 0;
std::string *flatlines;

const int NPAR = 8;

struct flattice
{
  std::string keyword;     // keyword MAD/SPINK char*4 element types
  std::string SWN;         // SiteWideName from NameLookup
  std::string swn;         // shortened SiteWideName without "-"'s
  std::string devname;     // DeviceName from NameLookup
  std::string latname;     // LatticeName from old MAD files
  std::string coordtype;   // CoordinateType from NameLookup
  double l;                // length of element [m]
  double sup;              // initial s-coordinate of element [m]
  double sdn;              // final s-coordinate of element [m]
  double par[NPAR];        // parameters for element (depends on keyword)
} *flt;
int nflat;

struct lattice
{
  std::string keyword;   // keyword MAD/SPINK char*4 element types
  std::string SWN;       // SiteWideName from NameLookup
  std::string swn;       // shortened SiteWideName without "-"'s
  std::string devname;   // DeviceName from NameLookup
  std::string latname;   // LatticeName from old MAD files
  std::string coordtype; // CoordinateType from NameLookup
  double s;              // longitudinal center of element
  double l;              // length of element
  double sup;            // upstream s-coord
  double sdn;            // downstream s-coord
  double par[NPAR];      // parameters for element (depends on keyword)
  Tfunc M;               // transport matrix and vector, ...
  int ithv;		 // =1 if hor corr (th), =2 for vert(tv), =0 otherwise.
  int spinflg;		 // =0 if elt is spin transparent; otherewise it's not.
  int spinfix;		 // !=0 if spin matrix does not need to be recalculated.
  su2mat smat;		 // spin matrix for element during present turn
  double xoff;           // save any hor misalignment for future ref (nl elts)
  double yoff;           // save any vert misalignment for future ref (nl elts)
} *lat;
int nelt = 0;  // number of elements filled in Tfunc lat[];
int nlatmax = 0;

struct bump
{
  std::string name;      // name of bump
  int n;		 // number of steering correctors in bump
  std::string *SWN;	 // array of SWN's for correctors
  double *coef;		 // corresponding coefficients (for 1mm amplitude)
  double amp;         // bump amplitude in meters
  int    bumpflg;	 // =0 if bump is off, otherwise it is on
} *bumps;
int nbumps;		 // number of bumps

struct snake
{
  std::string SWN;	 // SWN of snake
  double mu;		 // spin rotation angle (left-handed about nevec)
  double phi;		 // angle between z-axis and rotator axis (nevec)
  			 //   projected onto horiz plane
  double theta;		 // angle between rotator axis and horiz plane
  Trajectory error;	 // trajectory error through snake
} *snakes;
int nsnakes;		 // number of snakes and rotators

struct solen
{
  std::string SWN;	 // SWN of snake
  double B;		 // spin rotation angle (left-handed about nevec)
} *sols;
int nsol;		 // number of snakes and rotators

char drprefix[] = "dr"; // prefix for SWN drift name generation on the fly.


Tfunc M1t;
double totlen;  // calculated in fullturnmatrix()
double totang0, totangh, totangv, totangdt0;


char tab = '\t';
char cmsp[] = ", ";
char endl = '\n';
char whspc[] = ":\t, "; 
std::fstream fmat;
std::fstream fsm;
std::fstream fspinor;
int skipflg = 0; // flag >0 skips output to fsm and fspinor; gets decremented

FILE* ftmp;      // filehandle for a temporary file (debuggery)
int runnum;
std::string runnumfile = "blspin-runnum.dat";

//--- --- ---
//Stuff for stroboscopic averaging
struct polfield
{
  Trajectory X;  // 6d phase space point
  Point S;       // 3d spin vector
  su2spinor psi; // 2d complex spinor for spin 1/2
} *pf, *pfsum, *pf0;
int npfld = 0;
int lnvffile = 0;
int lnvtbtfile = 0;
FILE *tbtpffile = NULL;

//====================
int incrementrunnum(void)
{
  std::string tmp = strcat(getenv("USPASwwm"), "/blspin-runnum.dat");
  //  std::cout << tmp <<"\n";

  std::fstream fn;
  fn.open(runnumfile.c_str(), std::ios::in);
  if(fn==0)
    {
      std::cerr << "Could not open("<<runnumfile<<"); initializing runnum=0\n";
      runnum = 0;
    } else {
      fn >> runnum;
      fn.close();
    }
  fn.open(runnumfile.c_str(), std::ios::out);
  runnum++;
  fn << runnum << endl;
  fn.close();

  return runnum;
}
//---------------
void dumppar(const std::string name, double value)
{
  fprintf(fgdmp, "#  %s:\t%22.18e\n", name.c_str(), value);
}
//---------------
int process_cmd(int argc, char **argv)
{
  if(argc<2)
    {
      fprintf(stderr, "Usage: %s file.param\n", argv[0]);
      fprintf(stderr, " where file.param is the list of input parameters.\n");
      return -1;
    }
  GIfile = argv[1];
  nlin = 0; // assume linear optics unless changed in GIinput()
  GIinput();
  if(nbsteps<1) nbsteps = 1;
  if(nqsteps<1) nqsteps = 1;
  if(zerotol<=0) zerotol = 1.0e-12;
  GIdump();
  fprintf(fgdmp, "#zerotol:\t%12.6e\n", zerotol);
  fprintf(fgdmp, "# %s Run number: %d\n", argv[0], runnum);
  

  circumference = design_length;
  gam = U0/mc2;
  U0 *= 1.e9;    // convert to eV
  mc2 *= 1.e9;   // convert to eV
  beta = sqrt(1.0-1.0/gam/gam);
  rigidity = mc2*sqrt(gam*gam-1.0)/c;
  Gg   = G*gam;
  Ggp1 = Gg+1.0;
  Ggm1 = G*(gam-1.0);
  Gp1  = G+1.0;
  omegarf = twopi*beta*c*harmnum/circumference;

  dumppar("gam", gam);
  dumppar("U0", U0);
  dumppar("mc2", mc2);
  dumppar("beta", beta);
  dumppar("rigidity", rigidity);
  dumppar("Gg", Gg);
  dumppar("Ggp1", Ggp1);
  dumppar("Ggm1", Ggm1);
  dumppar("Gp1", Gp1);
  dumppar("omegarf", omegarf);

  if(lmatfile!=0)
    {
      fmat.open(matfile.c_str(), std::ios::out);
      if(fmat==0)
	{
	  std::cerr<<"Could not open("<<matfile<<").\n";
	  return -1;
	}
    }
  if(lsm!=0)
    {
      fsm.open(smfile.c_str(), std::ios::out);
      if(fsm==0)
	{
	  std::cerr<<"Could not open("<<smfile<<").\n";
	  return -1;
	}
    }
  if(lspinor!=0)
    {
      fspinor.open(spinorfile.c_str(), std::ios::out);
      if(fspinor==0)
	{
	  std::cerr<<"Could not open("<<spinorfile<<").\n";
	  return -1;
	}
    }

  return 0;
}
//-----------
void parseline(std::string str, struct flattice *flt, int i)
{
  int j;
  char line[MAXLIN];
  char* t;

  strncpy(line, str.c_str(), MAXLIN);
  t = strtok(line, "\t, ");
  flt[i].keyword = t;
  t = strtok(NULL, "\t, ");
  flt[i].SWN = t;
  t = strtok(NULL, "\t, ");
  flt[i].swn = t;
  t = strtok(NULL, "\t, ");
  flt[i].devname = t;
  t = strtok(NULL, "\t, ");
  flt[i].latname = t;
  t = strtok(NULL, "\t, ");
  flt[i].coordtype = t;

  t = strtok(NULL, "\t, ");
  flt[i].l = atof(t);
  t = strtok(NULL, "\t, ");
  flt[i].sup = atof(t);
  t = strtok(NULL, "\t, ");
  flt[i].sdn = atof(t);
  for(j=0;j<NPAR;j++)
    {
      t = strtok(NULL, "\t, ");
      flt[i].par[j] = atof(t);
    }
}
//-----------
void addelement(struct flattice flat, int j)
{
  int i, k;
  Point ne;
  double r;
  Trajectory kick;

  lat[j].keyword  = flat.keyword;
  lat[j].SWN      = flat.SWN;
  lat[j].swn      = flat.swn;
  lat[j].devname  = flat.devname;
  lat[j].latname  = flat.latname;
  lat[j].coordtype= flat.coordtype;
  lat[j].l   = flat.l;
  lat[j].sup = flat.sup;
  lat[j].sdn = flat.sdn;
  lat[j].s   = (flat.sup+flat.sdn)/2.;
  lat[j].ithv = 0;
  lat[j].spinflg = 0; // assume zero
  lat[j].spinfix = 0;
  for(i=0;i<NPAR;i++)
    {
      lat[j].par[i] = flat.par[i];
    }

  if(flat.keyword=="DRIF")
    {
      lat[j].M = Drift(lat[j].l, beta);
      return;
    }
  if(flat.keyword=="ECOL" || flat.keyword=="RCOL"
     || flat.keyword=="ELSE" || flat.keyword=="BEAM"
     || flat.keyword=="MATR" || flat.keyword=="LUMP"
     || flat.keyword=="YROT")
    {
      std::cerr << "addelement: "<< flat.SWN << " keyword: ";
      std::cerr << flat.keyword;
      std::cerr << " ignored.  Assuming a drift.\n";
      lat[j].keyword = "DRIF";
      return;
    }
  if(flat.keyword=="SBEN")
    {
      lat[j].M = Sbend(lat[j].l, flat.par[0], flat.par[1], flat.par[4],
		       flat.par[5], flat.par[3], 1, beta);
      lat[j].spinflg = 1;
      if(lat[j].devname=="ki") lat[j].spinflg = 0; // ignore injection kickers
      return;
    }
  if(flat.keyword=="RBEN")
    {
      double theta = flat.par[0];
      lat[j].M = Sbend(lat[j].l, theta, flat.par[1], theta/2.+flat.par[4],
		       theta/2.+flat.par[5], flat.par[3], 1, beta);
      lat[j].spinflg = 1;
      if(lat[j].devname=="ki") lat[j].spinflg = 0;
      return;
    }
  if(flat.keyword=="QUAD")
    {
      if(lat[j].latname=="qf" && kqf!=0)flat.par[1]=kqf;
      if(lat[j].latname=="qd" && kqd!=0)flat.par[1]=kqd;
      lat[j].M = Quad(flat.l, flat.par[1], flat.par[3], 1, beta);
      lat[j].spinflg = 1;
      return;
    }
  if(flat.keyword=="SEXT") // temporary kluge for linear lattice
    {
      lat[j].M = Drift(lat[j].l, beta);
      //      lat[j].spinflg = 1;  // temporarily set to zero
      return;
    }
  if(flat.keyword=="OCTU") // temporary kluge for linear lattice
    {
      lat[j].M = Drift(lat[j].l, beta);
      //      lat[j].spinflg = 1;  // temporarily set to zero
      return;
    }
  if(flat.keyword=="MULT")
    {  // considers only normal quadrupole for now
      lat[j].M.unit();
      lat[j].M.t.m[1][0] = -flat.par[1];
      lat[j].M.t.m[3][2] =  flat.par[1];
      if(flat.par[1]!=0) lat[j].spinflg = 1;
      return;
    }
  if(flat.keyword=="SOLE")
    {
      if(fabs(flat.par[4])<=zerotol)
	{
	  lat[j].M = Drift(lat[j].l, beta);
	  lat[j].spinflg = 1; // the solenoids get set later
	}
      else
	{
	  lat[j].M = Solen(lat[j].l, flat.par[4], beta);
	  lat[j].spinflg = 1;
	}
      return;
    }
  if(flat.keyword=="HKIC")
    {
      lat[j].M = Drift(lat[j].l, beta);
      lat[j].spinflg = 1;
      lat[j].ithv = 1;
      lat[j].spinfix = 1;  // temporary
      return;
    }
  if(flat.keyword=="VKIC")
    {
      lat[j].M = Drift(lat[j].l, beta);
      lat[j].spinflg = 1;
      lat[j].spinfix = 1;  // temporary
      return;
    }
  if(flat.keyword=="KICK")
    {
      lat[j].M = Drift(lat[j].l, beta);
      lat[j].spinflg = 1;
      lat[j].ithv = 1;
      std::cerr << "addelement: Found a KICK -- bailing.\n";
      exit(1);
      return;
    }
  if(flat.keyword=="RFCA")
    {
      //      lat[j].M = Drift(lat[j].l, beta);
      //      if(lat[j].SWN=="bi4-cavs3.1")
      if(lat[j].SWN==rfSWN)
	lat[j].M = Cavity(Vrf, Z, beta, mc2, omegarf, phirf);
      return;
    }
  if(flat.keyword=="INST")
    {
      if((lat[j].devname=="pol") || (lat[j].devname=="pjet"))
	{
	  lat[j].spinflg = 1;
	}
    }
  if(flat.keyword=="MARK")
    {
      if(flat.devname!="rot" && flat.devname!="snk")
	{
	  lat[j].spinflg = 1;
	}
    }
  if(flat.keyword=="SROT")
    {
      lat[j].M = Srot(flat.par[5], 1);
      std::cerr << "addelement: Found an SROT -- bailing.\n";
      exit(1);
      //      return;
    }
}
//-----------
int buildlattice(void)
{
  int i, ii, j, ielt, ndrif;
  double len;
  //  double sup, sdn, sprev, len;
  //  std::string kwd, SWN, swn, devnm, latnm, coordtype;
  char *line, tmpname[20];

  ndrif = 0;
  /*
  for(i=0;i<nlines;i++)
    {
      if(i<10) std::cout << flatlines[i];
      if(flatlines[i].substr(0,1)=="#")continue;  // skip comments
    }
  */
  nlatmax = 2*nlines+2;  // I'm always uncertain if the +? is needed here?
  flt = new struct flattice[nlatmax];
  lat = new struct lattice[nlatmax];

  ielt = 0;
  lat[0].keyword = "MARK";
  lat[0].SWN = "begin";
  lat[0].swn = "begin";
  lat[0].devname = "mark";
  lat[0].latname = "begin";
  lat[0].coordtype = "Traj";
  lat[0].l = 0;
  lat[0].sup = 0;
  lat[0].sdn = 0;
  lat[ielt].s = (lat[ielt].sup+lat[ielt].sdn)/2.;
  lat[0].M.unit();
  lat[0].spinflg = 0;
  ielt++;
  ii = 0;
  j = 0;
  for(i=0;i<nlines;i++)
    {
      if(flatlines[i].substr(0,1)=="#")continue;  // skip comments
      parseline(flatlines[i], flt, j);
      j++;
    }
  nflat = j;
  j = 0;
  for(i=0;i<nlines;i++)
    {
      if(flatlines[i].substr(0,1)=="#")continue;  // skip comments
      addelement(flt[j], ielt);
      j++;
      ielt++;
      ii++;
      len = flt[ii].sup-lat[ielt-1].sdn;
      //      std::cout <<"i="<<i<<"SWN:"<<flt[ii].SWN<<"\t len="<<len<<"\n";
      
      if(len!=0 && ii<nflat)
	{
	  lat[ielt].keyword = "DRIF";
	  sprintf(tmpname, "%s%d", drprefix, ndrif);
	  lat[ielt].SWN = tmpname;
	  lat[ielt].swn = tmpname;
	  lat[ielt].devname = "drift";
	  lat[ielt].latname = tmpname;
	  lat[ielt].coordtype = "Traj";
	  lat[ielt].sup = flt[ii-1].sdn;
	  lat[ielt].sdn = flt[ii].sup;
	  lat[ielt].l = len;
	  lat[ielt].s = (lat[ielt].sup+lat[ielt].sdn)/2.;
	  lat[ielt].M = Drift(lat[ielt].l, beta);
	  lat[ielt].spinflg = 0;
	  ielt++;
	  ndrif++;
	}
    }
  if(ielt>2*nlines+1)
    {
      std::cerr << "oops: ielt="<<ielt<<" for nlines="<<nlines<<".\n";
      return -1;
    }
  lat[ielt].keyword = "MARK";
  lat[ielt].SWN = "end";
  lat[ielt].swn = "end";
  lat[ielt].devname = "mark";
  lat[ielt].latname = "end";
  lat[ielt].coordtype = "Traj";
  lat[ielt].l = 0;
  lat[ielt].sup = lat[ielt-1].sdn;
  lat[ielt].sdn = lat[ielt].sup;
  lat[ielt].s = (lat[ielt].sup+lat[ielt].sdn)/2.;
  lat[ielt].M.unit();
  lat[ielt].spinflg = 0;
  
  nelt = ielt+1;

  return 0;
}
//----------------
int readflat(std::string fnam)
{ // read in data from the flat lattice file.  (Waldo's format)
  FILE *f;
  int i;
  char line[MAXLIN];
  
  f = fopen(fnam.c_str(), "r");
  if(f==NULL)
    {
      std::cerr << "Could not fopen("<<fnam<<",r).\n";
      return -1;
    }
  //Count the lines in the flat file.  Warning: Hopefully MAXLIN is big enough.
  while(fgets(line, MAXLIN, f))
    {
      nlines++;
      //      if(nlines<10)std::cout << nlines <<":\t"<<line<<"*\n";
    }
  std::cerr << "file: "<<fnam<<" has "<<nlines<<" lines in it.\n";

  rewind(f);
  flatlines = new std::string[nlines];
  
  i = 0;
  while(fgets(line, MAXLIN, f))
    {
      flatlines[i] = line;
      i++;
    }

  if(buildlattice()) return -1;

  delete [] flatlines;
  fclose(f);
  return 0;
}
//----------------
void applyoffset(const std::string name, const std::string xy,
		 const double off)
{  // deprecated function
  Trajectory TT;
  Tfunc MM;
  int iii;
  for(iii=0;iii<nelt;iii++)
    {
      if(lat[iii].SWN==name)
	{
	  if(xy=="x"||xy=="h")
	    {
	      TT.v[0] = off;
	      MM.t = MM.t-lat[iii].M.t;
	      lat[iii].M.v = MM*TT;
	      //     lat[iii].M.v.v[0] = off;
	      lat[iii].xoff = off;  // can be used to fix nonlin elts
	    }
	  else if(xy=="y"||xy=="v")
	    {
	      TT.v[2] = off;
	      MM.t = MM.t-lat[iii].M.t;
	      lat[iii].M.v = MM*TT;
	      //     lat[iii].M.v.v[2] = off;
	      lat[iii].yoff = off;  // can be used to fix nonlin elts
	    }
	  return;
	}
    }
}
//----------------
void applyoffsets(const std::string name, const double xoff,
		 const double yoff)
{
  Trajectory TT;
  Tfunc MM;
  int iii;
  for(iii=0;iii<nelt;iii++)
    {
      if(lat[iii].SWN==name)
	{
	  TT.v[0] = xoff;
	  TT.v[2] = yoff;
	  MM.t = MM.t-lat[iii].M.t;
	  lat[iii].M.v = MM*TT;
	  //     lat[iii].M.v.v[0] = off;
	  lat[iii].xoff = xoff;  // can be used to fix nonlin elts
	  lat[iii].yoff = yoff;  // can be used to fix nonlin elts
	  return;
	}
    }
}
//----------------
int readoffsets(const std::string fnam)
{
  FILE *f;
  std::string swn;
  double xoff, yoff;
  char line[MAXLIN];
  char *t;

  f = fopen(fnam.c_str(), "r");
  if(f==NULL)
    {
      //      std::cerr << "readoffsets: Could not fopen("<<fnam<<",r).\n";
      return -1;
    }
  
  while(fgets(line, MAXLIN, f))
  {
    if((line[0]=='\n') || (line[0]=='#'))continue;
    t = strtok(line, whspc);
    swn = t;
    t = strtok(NULL, whspc);
    xoff = scaleoffset*atof(t);
    t = strtok(NULL, whspc);
    yoff = scaleoffset*atof(t);
    //    applyoffset(swn, "x", xoff);
    //    applyoffset(swn, "y", yoff);
    applyoffsets(swn, xoff, yoff);
  }

  fclose(f);
  f = NULL;
  return 0;
}
//----------------
void dumpsnakelist(void)
{
  int i, j;

  fprintf(fgdmp, "#snake and rotator parameters:\n");
  fprintf(fgdmp, "#SWN, mu, phi, theta, dx, dx', dy, dy' dz\n");
  for(i=0;i<nsnakes;i++)
    {
      fprintf(fgdmp, "# %s\t%11.6f %11.6f %11.6f ",
	      snakes[i].SWN.c_str(),
	      snakes[i].mu, snakes[i].phi, snakes[i].theta);
      for(j=0;j<5;j++)
	{
	  fprintf(fgdmp, " %10.6f", snakes[i].error.v[j]);
	}
      fprintf(fgdmp, "\n");
    }
}
//----------------
int readsnake(std::string fnam)
{
  FILE *f;
  int i, j;
  char line[MAXLIN];
  char *t;
  Point ne;
  double r;
  

  f = fopen(fnam.c_str(), "r");
  if(f==NULL)
    {
      //      std::cerr << "Could not fopen("<<fnam<<",r).\n";
      return -1;
    }
  i = 0;
  while(fgets(line, MAXLIN, f))
    {
      if((line[0]=='\n') || (line[0]=='#'))continue;
      t = strtok(line, whspc);
      if(strcmp(t, "nsnakes")==0)
	{
	  if(nsnakes!=0)
	    {
	      std::cerr<<"Error: readsnake more than one 'nsnakes' line."
		       <<endl;
	      return -1; // bail out (we could ignore it I suppose.
	    }
	  t = strtok(NULL, whspc);
	  nsnakes = atoi(t);
	  snakes = new struct snake[nsnakes];
	  goto next;
	}
      if(i>=nsnakes) goto skipit;
      
      snakes[i].SWN = t;
      t = strtok(NULL, whspc);
      snakes[i].mu    = atof(t)*pi/180.0;
      t = strtok(NULL, whspc);
      snakes[i].phi   = atof(t)*pi/180.0;
      t = strtok(NULL, whspc);
      snakes[i].theta = atof(t)*pi/180.0;
      ne.y = sin(snakes[i].theta);
      r = cos(snakes[i].theta);
      ne.x = r*sin(snakes[i].phi);
      ne.z = r*cos(snakes[i].phi);

      snakes[i].error.zero();
      for(j=0;j<5;j++)
	{
	  t = strtok(NULL, whspc);
	  snakes[i].error.v[j] = atof(t);
	}
      for(j=0;j<nelt;j++)
	{
	  if(snakes[i].SWN==lat[j].SWN)
	    {
	      lat[j].spinflg = 1;
	      lat[j].spinfix = 1;
	      lat[j].M.v = snakes[i].error;
	      ne.y = sin(snakes[i].theta);
	      r = cos(snakes[i].theta);
	      ne.x = r*sin(snakes[i].phi);
	      ne.z = r*cos(snakes[i].phi);
	      lat[j].smat = Rn(snakes[i].mu, ne);
	      goto gotit;
	    }
	}
    gotit:
      i++;
    next: continue;
    }
 skipit:
  fclose(f);
  dumpsnakelist();
  return 0;
}
//----------------
void dumpbumplist(void)
{
  int i, j;

  fprintf(fgdmp, "#bump definitions\n");
  for(i=0;i<nbumps;i++)
    {
      fprintf(fgdmp, "#Bump: %s amplitude=%f [m]  bumpflg=%d "
	      ,bumps[i].name.c_str()
	      ,bumps[i].amp
	      ,bumps[i].bumpflg
	      );
      fprintf(fgdmp, "  SWN, coef\n");
      for(j=0;j<bumps[i].n;j++)
	{
	  fprintf(fgdmp, "#  %s\t%11.6f\n",
		  bumps[i].SWN[j].c_str(), bumps[i].coef[j]);
	}
    }
}
//----------------
int readbump(std::string fnam)
{
  FILE *f;
  int i, j;
  int nb, ib;
  char line[MAXLIN];
  char *t;

  nb = 0;  // why wasn't this initially zero?
  f = fopen(fnam.c_str(), "r");
  if(f==NULL)
    {
      std::cerr << "Could not fopen("<<fnam<<",r).\n";
      return -1;
    }
  i = -1;
  while(fgets(line, MAXLIN, f))
    {
      if((line[0]=='\n') || (line[0]=='#'))continue;
      t = strtok(line, whspc);
      if(strcmp(t, "nbumps")==0)
	{
	  if(nbumps!=0)
	    {
	      std::cerr<<"Error: readbump more than one 'nbumps' line."
		       <<endl;
	      return -1; // bail out (we could ignore it I suppose.
	    }
	  t = strtok(NULL, whspc);
	  nbumps = atoi(t);
	  bumps = new struct bump[nbumps];
	  goto next;
	}
      if(i>=nbumps)goto skipit;
      if(strcmp(t, "name")==0)
	{
	  i++;
	  t = strtok(NULL, whspc);
	  bumps[i].name = t;
	  t = strtok(NULL, whspc);
	  nb = atoi(t);
	  t = strtok(NULL, whspc);
	  bumps[i].amp = atof(t);
	  t = strtok(NULL, whspc);
	  bumps[i].bumpflg = atoi(t);
	  bumps[i].n = nb;
	  bumps[i].SWN  = new std::string[nb];
	  bumps[i].coef = new double[nb];
	  ib = 0;
	  goto next;
	}
      if(ib>=nb)
	{
	  std::cerr<<"readbumps: problem line=:"<<line;
	  return -1;
	}
      bumps[i].SWN[ib] = t;
      t = strtok(NULL, whspc);
      bumps[i].coef[ib] = atof(t);
      ib++;
    next: continue;
    }
 skipit:
  fclose(f);
  dumpbumplist();
  return 0;
}
//----------------
int readdBoB(std::string fdB)
{
  FILE *f;
  int j;
  char line[MAXLIN];
  char *t;
  std::string magname;
  double dbob;
  Tfunc M1, M2;
  mat6 I;
  
  Trajectory t1, t2;

  f = fopen(fdB.c_str(), "r");
  if(f==NULL)
    {
      std::cerr << "Could not fopen("<<fdB<<",r).\n";
      return -1;
    }
  while(fgets(line, MAXLIN, f))
    {
      if((line[0]=='\n') || (line[0]=='#'))continue;
      magname = strtok(line, whspc);
      t = strtok(NULL, whspc);
      dbob = atof(t);
      if(dbob==0.0) continue;
      for(j=0;j<nelt;j++)
	{
	  if(lat[j].SWN==magname)
	    {
	      M1 = lat[j].M;
	      t1.v[5] = -dbob;
	      M2.t = M1.t-I;
	      t2 = M2*t1;
	      lat[j].M.v += t2;
	      std::cerr<<"dBoB: "<<magname<<"\n";
	      std::cerr<<lat[j].M;
	      continue;
	    }
	}
    }

  return 0;
}
//----------------
void dumpsollist(void)
{
  int i, j;

  fprintf(fgdmp, "#sol definitions\n");
  for(i=0;i<nsol;i++)
    {
      fprintf(fgdmp, "#Solenoid %s B = %f [T]\n",
	      sols[i].SWN.c_str() ,sols[i].B);
    }
}
//----------------
int readsol(std::string fnam)
{
  FILE *f;
  int i, j;
  char line[MAXLIN];
  char *t;
  Point ne;
  double r;
  

  f = fopen(fnam.c_str(), "r");
  if(f==NULL)
    {
      //      std::cerr << "Could not fopen("<<fnam<<",r).\n";
      return -1;
    }
  i = 0;
  while(fgets(line, MAXLIN, f))
    {
      if((line[0]=='\n') || (line[0]=='#'))continue;
      t = strtok(line, whspc);
      if(strcmp(t, "nsol")==0)
	{
	  if(nsol!=0)
	    {
	      std::cerr<<"Error: readsol more than one 'nsol' lines."
		       <<endl;
	      return -1; // bail out (we could ignore it I suppose).
	    }
	  t = strtok(NULL, whspc);
	  nsol = atoi(t);
	  sols = new struct solen[nsol];
	  goto next;
	}
      if(i>=nsol) goto skipit;
      
      sols[i].SWN = t;
      t = strtok(NULL, whspc);
      sols[i].B    = atof(t);

      for(j=0;j<nelt;j++)
	{
	  if(sols[i].SWN==lat[j].SWN)
	    {
	      lat[j].M = Solen(lat[j].l, sols[i].B/rigidity, beta);
	      continue;
	    }
	}
      i++;
    next: continue;
    }
 skipit:
  fclose(f);
  dumpsollist();
  return 0;
}
//----------------
void orbsteer(const int i, const double steer)
{
  std::cout<<"orbsteer("<<i<<", "<<lat[i].SWN<<", "<<lat[i].devname<<") "
	   <<steer<<endl;
  if(lat[i].devname=="th")
    {
      lat[i].M.v.v[1] += steer;
      lat[i].smat = Ry(-Ggp1*atan(lat[i].M.v.v[1]));
    }
  else if(lat[i].devname=="tv")
    {
      lat[i].M.v.v[3] += steer;
      lat[i].smat = Rx(Ggp1*atan(lat[i].M.v.v[3]));
    }
}
//----------------
void applybump(const int i, const double stren)
{
  int j, k;
  double steer;

  for(j=0;j<bumps[i].n;j++)
    {
      steer = stren*bumps[i].coef[j];
      for(k=0;k<nelt;k++)
	{
	  if(lat[k].SWN==bumps[i].SWN[j])
	    {
	      orbsteer(k, steer);
	      goto nextcor;
	    }
	}
    nextcor: continue;
    }
}
//----------------
void applybump(const std::string bnm, const double stren)
{
  int i, j, k;

  for(i=0;i<nbumps;i++)
    {
      if(bumps[i].name==bnm)
	{
	  applybump(i, stren);
	}
    }
}
//----------------
void dumplat(void)
{
  int i;

  fprintf(fgdmp, "#nelt =%d\n", nelt);
  for(i=0;i<nelt;i++)
    {
      fprintf(fgdmp, "%5d %4s %15s %15s %15s %14.8f %14.8f %14.8f %14.8f\n",
	      i, lat[i].keyword.c_str(), lat[i].SWN.c_str(),
	      lat[i].devname.c_str(),
	      lat[i].latname.c_str(),
	      lat[i].l, lat[i].sup, lat[i].sdn, lat[i].s);
    }
}
//----------------
void cleanup(void)
{
  int i;

  delete [] lat;
  delete [] flt;

  delete [] snakes;
  delete [] sols;
  for(i=0;i<nbumps;i++)
    {
      delete [] bumps[i].SWN;
      delete [] bumps[i].coef;
    }

  delete [] pf0;
  delete [] pf;
  delete [] pfsum;
}
//---------------
#include "nl_blspin-spin-inc.cxx"
//---------------
Tfunc linfullturnmatrix(void)
{ // Note nonlinear elements were initialized with a unit or drift matrix.
  // This ignores any nonlinear effects for this calculation.
  int i;
  Tfunc MM, t;
  for(i=0;i<nelt;i++)
    {
      t = lat[i].M*MM;
      MM = t;
      totlen += lat[i].l;
    }

  return MM;
}
//----------------
Point nevec(const su2mat R)
{
  double Vmag;
  Point V;
  V.z = imag(R.m[0][0]);
  V.x = imag(R.m[0][1]);
  V.y = real(R.m[0][1]);
  Vmag = V*V;
  if(Vmag>0)
    {
      V = V/sqrt(Vmag);
    }
  else
    {
      V.y = 1.0;
    }
  return V;
}
//----------------
double muof(const su2mat R)
{
  return acos(real(R.m[0][0]))*2*180.0/pi;
}
//----------------
void dumpspinmat(const int i, const su2mat t, const su2mat RR)
{
  char angle[15];
  double aa;

  mat3 m, rr;

  if(skipflg>0) return;


  m.m[0][0] = real(t.m[0][0]*t.m[0][0]-t.m[0][1]*t.m[0][1]);
  m.m[0][1] = imag(t.m[0][0]*t.m[0][0]+t.m[0][1]*t.m[0][1]);
  m.m[1][0] =-imag(t.m[0][0]*t.m[0][0]-t.m[0][1]*t.m[0][1]);
  m.m[1][1] = real(t.m[0][0]*t.m[0][0]+t.m[0][1]*t.m[0][1]);
  m.m[2][2] = real(t.m[0][0]*conj(t.m[0][0])-t.m[0][1]*conj(t.m[0][1]));
  m.m[0][2] =-2*real(t.m[0][0]*t.m[0][1]);
  m.m[1][2] = 2*imag(t.m[0][0]*t.m[0][1]);
  m.m[2][0] = 2*real(t.m[0][0]*conj(t.m[0][1]));
  m.m[2][1] = 2*imag(t.m[0][0]*conj(t.m[0][1]));

  rr.m[0][0] = real(RR.m[0][0]*RR.m[0][0]-RR.m[0][1]*RR.m[0][1]);
  rr.m[0][1] = imag(RR.m[0][0]*RR.m[0][0]+RR.m[0][1]*RR.m[0][1]);
  rr.m[1][0] =-imag(RR.m[0][0]*RR.m[0][0]-RR.m[0][1]*RR.m[0][1]);
  rr.m[1][1] = real(RR.m[0][0]*RR.m[0][0]+RR.m[0][1]*RR.m[0][1]);
  rr.m[2][2] = real(RR.m[0][0]*conj(RR.m[0][0])-RR.m[0][1]*conj(RR.m[0][1]));
  rr.m[0][2] =-2*real(RR.m[0][0]*RR.m[0][1]);
  rr.m[1][2] = 2*imag(RR.m[0][0]*RR.m[0][1]);
  rr.m[2][0] = 2*real(RR.m[0][0]*conj(RR.m[0][1]));
  rr.m[2][1] = 2*imag(RR.m[0][0]*conj(RR.m[0][1]));

  fsm<<i<<tab<<lat[i].keyword<<tab<<lat[i].SWN<<tab
     <<lat[i].l<<tab<<lat[i].par[0]<<endl;

  aa = muof(t);
  sprintf(angle, "%10.6f", aa);
  fsm<<t<<"\tn_e "<<lat[i].sup<<" :\t"<<nevec(t)
     <<tab<<"mu= "<<angle<<" deg"<<tab<<lat[i].SWN<<endl;
  fsm<<m;

  fsm<<"  cumulative:\n"<<RR;

  aa = muof(RR);
  sprintf(angle, "%10.6f", aa);
  fsm<<"\tn_c "<<lat[i].sup<<" :\t"<<nevec(RR)
     <<tab<<"mu= "<<angle<<" deg"<<tab<<lat[i].SWN<<endl;

  fsm<<rr;
}
//----------------
su2mat oneturnspin(Trajectory &Xout, const Trajectory Xin, const su2spinor Psi0)
{
  int i, j;
  su2mat RR, t;
  Trajectory XX = Xin;
  su2spinor Psi = Psi0;
  double k;
  totang0 = 0;
  totangdt0 = 0;
  totangh = 0;
  totangv = 0;

  std::string ftmpfile = "thetadump.dat";
  if(ftmp!=NULL)
    {
      fclose(ftmp);
      ftmp = NULL;
    }
  ftmp = fopen(ftmpfile.c_str(), "w");
  if(ftmp==NULL)
    {
      std::cerr<<"Could not fopen("<<ftmpfile<<")\n";
      exit(1);
    }
  
  RR.unit();
  if(lspinor)
    {  // Output a header line for the file to define the columns:
      fspinor << "# SWN, S, Sx, Sy, Sz, ";
      fspinor << "Psi[0].real, Psi[0].imag, Psi[1].real, Psi[1].imag, ";
      fspinor << " x, x', y, y', z, delta\n";
      fspinor << "#  1   2   3   4   5  ";
      fspinor << "6            7            8            9            ";
      fspinor << "10 11  12 13  14   15\n";
    }
  if(lspinor) dumpspinor(0, Psi, XX);
  for(i=0;i<nelt;i++)
    {
      if(lat[i].spinflg)
	{
	  //	  std::cout<<lat[i].SWN<<endl;
	  if(lat[i].keyword=="SBEN")
	    {
	      t = Sdip(i, XX);
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      XX = lat[i].M*XX;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="QUAD")
	    {
	      t = Squad(i, XX);
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      XX = lat[i].M*XX;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="SEXT")
	    {
	      if(nlin==0)
		{
		  t = Sstraight(i, XX);
		  XX = lat[i].M*XX;
		}
	      else
		{
		  t = Ssextipole(i, &XX);
		}
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="MULT")
	    {
	      if(nlin==0)
		{
		  t = Sstraight(i, XX);
		  XX = lat[i].M*XX;
		}
	      else
		{
		  t = Smultipole(i, &XX);
		}
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="HKIC")
	    {
	      t = Sstraight(i, XX);
	      //	      totangh += -atan(lat[i].M.v.v[1]);
	      //	      t = Ry(-Ggp1*atan(lat[i].M.v.v[1]));
	      //if(lat[i].M.v.v[1]!=0)
	      //{
	      //  std::cout<<lat[i].SWN<<": theta = "<<lat[i].M.v.v[1]<<endl;
	      //  std::cout<<t;
	      //}
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      XX = lat[i].M*XX;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="VKIC")
	    {
	      t = Sstraight(i, XX);
	      //	      totangv += atan(lat[i].M.v.v[3]);
	      //	      t = Rx(Ggp1*atan(lat[i].M.v.v[3]));
	      RR = t*RR;
	      lat[i].smat = t;
	      Psi = t*Psi;
	      XX = lat[i].M*XX;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if(lat[i].keyword=="SOLE")
	    {
	      t = Rz(0); // kluge
	      if(nsol>0)
		{
		  for(j=0;j<nsol;j++)
		    {
		      if(lat[i].SWN==sols[j].SWN && sols[j].B!=0.0)
			{
			  //	  t = Rz(lat[i].l*sols[j].B*(1+G)/rigidity
			  //		 /(1.+lat[i].M.v.v[5]));
			  k = sols[j].B/rigidity/(1+lat[i].M.v.v[5]);
			  t = Ssolen(lat[i].l, k, XX);
			  continue;
			}
		    }
		}
	      RR = t*RR;
	      Psi = t*Psi;
	      XX = lat[i].M*XX;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  else if((lat[i].devname=="pol") || (lat[i].devname=="mark")
		  ||(lat[i].devname=="pjet"))
	    {
	      if(lspinor) dumpspinor(i, Psi, XX);
	    }
	  else if(lat[i].devname=="snk" || lat[i].devname=="rot")
	    {
	      t = lat[i].smat;
	      RR =t*RR;
	      Psi = t*Psi;
	      if(lspinor) dumpspinor(i, Psi, XX);
	      if(lsm) dumpspinmat(i, t, RR);
	    }
	  if(std::isnan(real(t.m[1][1])))
	    {
	      std::cerr<<"oneturnspin: "<<"ielt="<<i<<"  "<<lat[i].SWN
		       <<"\n"<<t;
	      std::cerr<<"Particle lost?\n";
	      std::cerr<<XX;
	      exit(-1);
	    }
        }
      else
	{
	  if(lsm) dumpspinor(i, Psi, XX);
	  XX = lat[i].M*XX;
	}
    }
  if(lstrobave==0)
    {
      fprintf(stdout, "#oneturnspin: totang0   = %15.12f\n", totang0);
      fprintf(stdout, "#             totangdt0 = %15.12f\n", totangdt0);
      fprintf(stdout, "#             totangh   = %15.12f\n", totangh);
      fprintf(stdout, "#             totangv   = %15.12f\n", totangv);
      fprintf(fgdmp, "#oneturnspin: totang0 = %15.12f\n", totang0);
      fprintf(fgdmp, "#           totangdt0 = %15.12f\n", totangdt0);
      fprintf(fgdmp, "#             totangh = %15.12f\n", totangh);
      fprintf(fgdmp, "#             totangv = %15.12f\n", totangv);
    }
  Xout = XX;
  return RR;
}
//----------------
void dumpelt(struct lattice ELT)
{
  fmat <<ELT.keyword<<tab<< ELT.SWN <<tab<<ELT.s<<endl;
  fmat <<ELT.M;
}
//----------------
void dumponeturnM(void)
{
  double Qx, Qy;

  std::cout << "One turn matrix with totlen=" << totlen <<":\n";
  std::cout << M1t;
  Qx = acos((M1t.t.m[0][0]+M1t.t.m[1][1])/2.)/twopi;
  if(M1t.t.m[0][1]<0) Qx = 1.0-Qx;
  Qy = acos((M1t.t.m[2][2]+M1t.t.m[3][3])/2.)/twopi;
  if(M1t.t.m[2][3]<0) Qy = 1.0-Qy;
  std::cout << "Qx = "<< Qx << "\n";
  std::cout << "Qy = "<< Qy << "\n";
  std::cout << "beta_x = " << M1t.t.m[0][1]/sin(twopi*Qx) << "\n";
  std::cout << "beta_y = " << M1t.t.m[2][3]/sin(twopi*Qy) << "\n";
}
//----------------
Trajectory periorbit(Tfunc M)
{ // Given a full turn transfer map M, this solves the equation
  //   M.t * x + M.v = x + Delta
  // for the closed orbit where transpose(Delta) is the vector (0,0,0,0,DL,0)
  // with DL = circumference - design_length.

  Trajectory Delta, Xc, V;
  //  Delta.v[4] = circumference - totlen;
  Delta.v[4] = totlen - circumference;

  Tfunc tmp; // initialized with unit matrix and zeroed vector
  int i, j;

  for(i=0;i<TFDIM;i++)
    {
      for(j=0;j<TFDIM;j++)
        {
          tmp.t.m[i][j] -= M.t.m[i][j];
        }
      V.v[i] = M.v.v[i] - Delta.v[i];
    }
  tmp.t = inverse(tmp.t);
  //  std::cout << "inverse(I-M):\n" << tmp;
  Xc = tmp*V;
  // Xc now has tmp.t = (I-M.t) and tmp.v = (V-Delta)

  return Xc;
}
//----------------
void dumpturn(Trajectory Xc)
{
  FILE *forb;
  int i;
  Trajectory X, XX;
  double s;

  forb = fopen(orbitfile.c_str(), "w");
  if((forb=fopen("orbit.dat", "w"))==NULL)
    {
      std::cerr << "Could not fopen("<<orbitfile<<").\n";
    }

  fprintf(forb, "# s x xp y yp z delta\n");
  X = Xc;
  s = 0;
  fprintf(forb,
	  "%16.9f %12.9f  %12.9f  %12.9f  %12.9f  %12.9f  %12.9f %-10s %s\n",
	  s, X.v[0], X.v[1], X.v[2], X.v[3], X.v[4], X.v[5],
	  "init", "init");

  Tfunc MM;
  for(i=0;i<nelt;i++)
    {
      XX = lat[i].M*X;
      if(lmatfile)
	{
	  dumpelt(lat[i]);
	  //	  MM = lat[i].M*MM;
	  //	  fmat<<"cumulative matrix:\n"<<MM;
	}
      X = XX;
      s += lat[i].l;
      fprintf(forb,
      "%16.9f %12.9f  %12.9f  %12.9f  %12.9f  %12.9f  %12.9f %-10s  %s\n",
	      s, X.v[0], X.v[1], X.v[2], X.v[3], X.v[4], X.v[5],
	      lat[i].devname.c_str(), lat[i].SWN.c_str());
    }
}
//----------------
int dumpcorrectors(void)
{
  int i;
  FILE *fcor;
  fcor = fopen(corrfile.c_str(), "w");
  if(fcor==NULL)
    {
      std::cerr<<"Could not fopen("<<corrfile<<")\n";
      return -1;
    }

  fprintf(fcor, "# SiteWideName DevName Scoord thetax thetay\n");
  for(i=0;i<nelt;i++)
    {
      if(lat[i].devname=="th")
	{
	  fprintf(fcor, "%-15s %-5s %13.6f  %13.10f   %13.10f\n",
		  lat[i].SWN.c_str(), lat[i].devname.c_str(),
		  lat[i].s, lat[i].M.v.v[1], 0.0);
	}
      else if(lat[i].devname=="tv")
	{
	  fprintf(fcor, "%-15s %-5s %13.6f  %13.10f   %13.10f\n",
		  lat[i].SWN.c_str(), lat[i].devname.c_str(),
		  lat[i].s, 0.0, lat[i].M.v.v[3]);
	}
    }
  fclose(fcor);
}
//----------------
int readnvi(std::string ifil)
{
  FILE *f;
  int i;
  Point t;
  char line[MAXLIN];

  if(npart<1) return 0; // skip if no particles given.
  
  f = fopen(ifil.c_str(), "r");
  if(f==NULL)
    {
      std::cerr<<"Could not fopen("<<ifil<<")\n";
      return -1;
    }
  pf0 = new struct polfield[npart];    // initial spin field
  pf = new struct polfield[npart];     // spin field -- changes with turn
  pfsum = new struct polfield[npart];  // cumulative sum of pf over turns
  i = 0;
  while(fgets(line, MAXLIN, f))
    {
      if(line[0]=='#' || line[0]=='\n') continue;
      sscanf(line, "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
	     &pf0[i].X.v[0], &pf0[i].X.v[1], &pf0[i].X.v[2],
	     &pf0[i].X.v[3], &pf0[i].X.v[4], &pf0[i].X.v[5],
	     &t.x, &t.y, &t.z);
	//     &pf0[i].S.x, &pf0[i].S.y, &pf0[i].S.z);
	pf0[i].S = normalize(t);
      /*  Debuggery
      printf("%f %f %f %f %f %f %f %f %f\n",
	    pf0[i].X.v[0], pf0[i].X.v[1], pf0[i].X.v[2],
	    pf0[i].X.v[3], pf0[i].X.v[4], pf0[i].X.v[5],
	    pf0[i].S.x, pf0[i].S.y, pf0[i].S.z);
      */
      pf0[i].psi = su2spinor(pf0[i].S);
      pf[i].X = pf0[i].X;
      pf[i].S = pf0[i].S;
      pf[i].psi = pf0[i].psi;
      pfsum[i].X = pf0[i].X;
      pfsum[i].S = pf0[i].S;
      pfsum[i].psi = pf0[i].psi;
      i++;
      if(i>=npart) break;
    }
  npfld = i;
  fclose(f);
  f = NULL;

  return 0;
}
//----------------
int outputnvtbtfile(int part, int turn)
{ // output stroboscopic average for part particle and turn turn
  int i, j;
  //  double rn = 1.0;
  Point t;

  if(lnvtbtfile==0) return 0;

  if(tbtpffile==NULL)
    {
      tbtpffile = fopen(nvtbtfile.c_str(), "w");
      if(tbtpffile==NULL)
	{
	  std::cerr<<"Could not fopen("<<nvtbtfile<<").\n";
	  return -1;
	}
      fprintf(tbtpffile, "# turn particle x0, x0', y0, y0', z0, delta0,  ");
      fprintf(tbtpffile, "x, x', y, y', z, delta,  ");
      fprintf(tbtpffile, "Sx,  Sy, Sz,  ");
      fprintf(tbtpffile, "sumSx, sumSy, sumSz,  ");
      fprintf(tbtpffile, "strobSx, strobSy, strobSz\n");
      fprintf(tbtpffile, "#  1      2      3   4    5   6    7      8    ");
      fprintf(tbtpffile, "9 10  11 12  13    14    ");  
      fprintf(tbtpffile, "15   16  17   ");
      fprintf(tbtpffile, "   18     19     20   ");
      fprintf(tbtpffile, "     21       22       23\n");
    }

  //  if(turn>0) rn = 1.0/(double)turn;

  fprintf(tbtpffile, "%8d %8d ", turn, part);
  for(j=0;j<6;j++) fprintf(tbtpffile, "%15.9e ", pf0[part].X.v[j]);
  for(j=0;j<6;j++) fprintf(tbtpffile, "%15.9e ", pf[part].X.v[j]);
  fprintf(tbtpffile, "%15.9e %15.9e %15.9e",
	  pf[part].S.x, pf[part].S.y, pf[part].S.z);
  t = pfsum[part].S;
  fprintf(tbtpffile, "%15.7e %15.7e %15.7e", t.x, t.y, t.z);
  t = normalize(t);
  fprintf(tbtpffile, " %15.9e %15.9e %15.9e\n", t.x, t.y, t.z);
  
  return 0;
}
//----------------
int outputnvffile(void)
{ // output final nfield from stroboscopic avg.
  int i, j;
  FILE *f;
  double rn = (double) npfld;
  Point nvec;

  f = fopen(nvffile.c_str(), "w");
  if(f==NULL)
    {
      std::cerr<<"Could not fopen("<<nvffile<<")\n";
      return -1;
    }
  fprintf(f, "# turn x x' y y' z delta nx ny nz\n");
  for(i=0;i<npfld;i++)
    {
      fprintf(f, "%8d ", i);
      for(j=0;j<6;j++)
	fprintf(f, "%15.9f ", pf0[i].X.v[j]);
      nvec = normalize(pfsum[i].S);
      fprintf(f, " %15.9f %15.9f %15.9f\n",
	      nvec.x/rn, nvec.y/rn, nvec.z/rn);
    }
  fclose(f);
  f = NULL;
  return 0;
}
//----------------
int pftrack()
{
  int i, j, k;
  Trajectory Xout, X;
  su2spinor psi;
  Point sumS, Ssa;
  su2mat otsm;
  FILE *ftbt;

  if(lstrobave==0)return 0;
  skipflg = -1;
  if(lnvtbtfile!=0)
    {
      ftbt = fopen(nvtbtfile.c_str(), "w");
      if(ftbt==NULL)
	{
	  std::cerr<<"Could not fopen("<<nvtbtfile<<")\n";
	  return -1;
	}
    }

  /*
  // Set up some quick debuggery
  FILE *fdebtmp;
  std::complex<double> otsmdet;
  fdebtmp = fopen("ostm-det.dat", "w");
  if(fdebtmp == NULL)
    {
      std::cerr<<"Could not fopen(ostm-det.dat)\n";
      return -1;
    }
  */
  for(i=0;i<npfld;i++)
    {
      // track ith particle's spin and trajectory.
      X = pf[i].X;
      psi = pf[i].psi;
      for(j=0;j<nturns;j++)
	{
	  otsm = oneturnspin(Xout, pf[i].X, pf[i].psi);
	  /*
	  otsmdet = otsm.m[0][0]*otsm.m[1][1]-otsm.m[0][1]*otsm.m[1][0];
	  fprintf(fdebtmp, "%6d, %6d, %21.18e %21.18e\n",
		  i, j, real(otsmdet), imag(otsmdet));
	  */
	  pf[i].psi = otsm*pf[i].psi;
	  //  I noticed that the spinor became slightly unnorm'd
	  pf[i].psi = normalize(pf[i].psi);

	  pf[i].X = Xout;
	  pf[i].S = vector(pf[i].psi);
	  pfsum[i].S = pfsum[i].S + pf[i].S;
	  if(lnvtbtfile!=0)outputnvtbtfile(i, j);
	}
      fprintf(tbtpffile, "\n");
    }
  fclose(ftbt);
  ftbt = NULL;
  if(outputnvffile()) return -1;
  return 0;
}
//----------------
int readdata(int argc, char** argv)
{ // Read in the lattice and data configuration files
  int i;

  if(incrementrunnum()<0)
    {
      std::cerr<<"Problem incrementing runnum file\n";
      return -1;
    }

  if(process_cmd(argc, argv)) return -1;

  if(readflat(flatname)) return -1;
  if(snakefile!="")
    {
      if(readsnake(snakefile)) return -1;
    }
  if(bumpfile!="")
    {
      if(readbump(bumpfile)) return -1;
    }
  if(dBoBfile!="")
    {
      if(readdBoB(dBoBfile)) return -1;
    }
  dumplat();

  // try to debug the closed orbit calculation
  //  Trajectory q1off;   // offset first quad of "fodo.lat"
  //  q1off.v[0] = 0.001;
  //  lat[1].M.offset(q1off);

  if(scaleoffset!=0.0 && offsetfile!="")
    {
      if(readoffsets(offsetfile)) return -1;
    }
  for(i=0;i<nbumps;i++)
    {
      if(bumps[i].bumpflg!=0)
	{
	  applybump(i, bumps[i].amp);
	}
    }

  if(solfile!="")
    {
      if(readsol(solfile)) return -1;
    }
  if(nvifile!="")
    {
      if(readnvi(nvifile))return -1;
      if(nvffile!="") lnvffile = 1;
      if(nvtbtfile!="") lnvtbtfile = 1;
    }
  return 0;
}
//----------------
int main(int argc, char** argv)
{
  int i, j;
  char numstr[100];
  double nu0;

  //- - - - - -
  if(readdata(argc, argv)!=0) return -1;
  //- - - - - - - -
  // First calculate the 1-turn matrix with linear optics
  M1t = linfullturnmatrix();
  dumponeturnM();

  Trajectory Xc, X1;
  if(clorb==0)
    {
      Xc.v[0] = x_init;
      Xc.v[1] = xp_init;
      Xc.v[2] = y_init;
      Xc.v[3] = yp_init;
      Xc.v[4] = z_init;
      Xc.v[5] = delta_init;
      std::cout << "Trajectory:\n" << Xc;
      skipflg = 0; // =1 for only writing out spinor data for closed orbit
    }
  else
    {
      Xc = periorbit(M1t);
      //      std::cout << "Closed trajectory:\n" << Xc;
      printf("Closed Trajectory:\n");
      printf("%15.12f  %15.12f  %15.12f  %15.12f  %15.12f  %15.12f\n",
	 Xc.v[0], Xc.v[1], Xc.v[2], Xc.v[3], Xc.v[4], Xc.v[5]);
      skipflg = 1;
    }
  //  std::cout << "After 1 turn:\n" << M1t*Xc;
  printf("After 1 turn with linear optics:\n");
  X1 = M1t*Xc;
  printf("%15.12f  %15.12f  %15.12f  %15.12f  %15.12f  %15.12f\n",
	 X1.v[0], X1.v[1], X1.v[2], X1.v[3], X1.v[4], X1.v[5]);
  printf("Excess path length: %15.12f\n", X1.v[4]-Xc.v[4]);
  dumpturn(Xc);

  Trajectory Xout;
  Point n_psi;
  n_psi = Point(Sxin, Syin, Szin);
  n_psi = normalize(n_psi);
  if(helicity==1)
    {
      n_psi = Point(xp_init, yp_init, 1);
      n_psi = normalize(n_psi);
    }
  if(clorb==0)
    {
      sprintf(numstr, "n_psi=[ %15.12f;  %15.12f;  %15.12f ]",
	      n_psi.x, n_psi.y, n_psi.z);
      std::cout<<"initial "<<numstr<<endl;
    }
  su2spinor psi, psi0(n_psi);
  //  std::cout <<"psi0:\n"<<psi0;

  su2mat Sturn0 = oneturnspin(Xout, Xc, psi0);
  skipflg--;

  Xout = M1t*Xc;
  std::cout <<"Xout:\n"<<Xout;
  Point phat = normalize(Point(Xout.v[1], Xout.v[3], 1.0));
  sprintf(numstr, "[ %21.18f %21.18f %21.18f ]", phat.x, phat.y, phat.z);
  std::cout <<"phat: [ "<<numstr<< " ]\n";
  
  std::cout <<"one turn spin rotation matrix:\n"<<Sturn0;
  if(clorb)
    std::cout <<"n0vec:\t"<<nevec(Sturn0)<<endl;
  else
    std::cout <<"rot axis:\t"<<nevec(Sturn0)<<endl;

  nu0 = acos(real(Sturn0.m[0][0]))/pi;
  sprintf(numstr, "%15.12f", nu0);
  std::cout <<"nu0:  \t"<<numstr<<endl;

  if(clorb)
    {
      psi0 = nevec(Sturn0);
      Sturn0 = oneturnspin(Xout, Xc, psi0);
      std::cout <<"psi0:\n"<<psi0;
    }
  psi = Sturn0*psi0;

  std::cout <<"psi_f:\n"<<psi<<endl;
  Point nf = vector(psi);
  sprintf(numstr, "[ %21.18f %21.18f %21.18f ]", nf.x, nf.y, nf.z);
  std::cout <<"final spinor vector: "<<numstr;
  std::cout <<"  norm^2-1 check: "
	    << (nf.x*nf.x+nf.y*nf.y+nf.z*nf.z)-1.0<<endl;
  sprintf(numstr, "%21.18f", nf*phat);
  std::cout <<"  nf.phat = "<<numstr<<endl;

  sprintf(numstr, "%15.12f", nu0);
  if(clorb)
    {
      std::cout <<"Closed orbit spin tune nu0:  \t"<<numstr<<endl;
      fprintf(fgdmp, "SPIN TUNE: nu_0 = %s\n", numstr);
    }
  else
    {
      std::cout <<endl<<"closed orbit not calculated (clorb="
		<<clorb<<")"<<endl;
      std::cout <<endl<<"1-turn spin rotation angle mu:\t"<<numstr<<endl;
      fprintf(fgdmp, "1-turn spin rotation angle mu: nu_0 = %s\n", numstr);
    }

  if(lcorrdmp)
    {
      if(dumpcorrectors()) return -1;
    }

  if(pftrack())return -1;

  cleanup();
  return 0;
}
